import numpy

from cofi import BaseProblem


_dummy_dist = lambda m: -numpy.inf if (m[0]<0 or m[0]>1) else 0
_dummy_dist_with_blobs = lambda m: (_dummy_dist(m), m, 2*m)
_blobs_dtype = [("model", float), ("model_doubled", float)]

def test_prior_likelihood():
    inv_problem = BaseProblem()
    inv_problem.set_log_prior(_dummy_dist)
    inv_problem.set_log_likelihood(_dummy_dist)
    assert inv_problem.log_prior_defined
    assert inv_problem.log_likelihood_defined
    assert inv_problem.log_posterior_defined

def test_prior_likelihood_skip_ll():
    inv_problem = BaseProblem()
    inv_problem.set_log_prior(_dummy_dist)
    def _failing_dist(_): assert "this line shouldn't be reached"
    inv_problem.set_log_likelihood(_failing_dist)
    inv_problem.log_posterior(numpy.array([-1]))

def test_posterior():
    inv_problem = BaseProblem()
    inv_problem.set_log_posterior(_dummy_dist)
    assert not inv_problem.log_prior_defined
    assert not inv_problem.log_likelihood_defined
    assert inv_problem.log_posterior_defined

def test_posterior_with_blobs():
    inv_problem = BaseProblem()
    inv_problem.set_log_posterior_with_blobs(_dummy_dist_with_blobs, _blobs_dtype)
    assert not inv_problem.log_prior_defined
    assert not inv_problem.log_likelihood_defined
    assert inv_problem.log_posterior_defined
    assert inv_problem.log_posterior_with_blobs_defined
    assert inv_problem.blobs_dtype_defined



from numpy import argmin
from numpy.typing import NDArray
from typing import Any

from . import BaseInferenceTool, error_handler


class NeighpyI(BaseInferenceTool):
    r"""Wrapper for the tool :code:`neighpy`, a Python implementation of the Neighbourhood Algorithm.

    The Neighbourhood Algorithm is a direct search method that uses a set of points in the parameter space to explore the objective function. It is a derivative-free method that is particularly useful for high-dimensional problems.

    This wrapper only implements the 1st phase of the NA i.e. the direct search phase.
    """

    documentation_links = ["https://neighpy.readthedocs.io/en/latest/index.html"]
    short_description = [
        "Wrapper for Neighpy, a Python implementation of the Neighbourhood Algorithm"
    ]

    @classmethod
    def required_in_problem(cls) -> set:
        return {"objective"}

    @classmethod
    def optional_in_problem(cls) -> dict:
        return {}

    @classmethod
    def required_in_options(cls) -> set:
        return {
            "bounds",
            "n_samples_per_iteration",
            "n_cells_to_resample",
            "n_initial_samples",
            "n_iterations",
        }

    @classmethod
    def optional_in_options(cls) -> dict:
        return {"serial": False}

    def __init__(self, inv_problem, inv_options) -> None:
        super().__init__(inv_problem, inv_options)
        self._params["ndim"] = len(self._params["bounds"])
        self._components_used = list(self.required_in_problem())

    def __call__(self) -> dict:
        raw_results = self._call_backend_tool()
        _best = raw_results["samples"][argmin(raw_results["objectives"])]
        res = {"success": True, "model": _best}
        return {**res, **raw_results}

    @error_handler(
        when="at some point in the Neighbourhood Algorithm Direct Search phase",
        context="",
    )
    def _call_backend_tool(self) -> dict:
        searcher = self._initialise_searcher()
        samples, objectives = self._call_searcher(
            searcher, parallel=not self._params["serial"]
        )
        return {
            "samples": samples,
            "objectives": objectives,
        }

    @staticmethod
    @error_handler(
        when="in calling neighpy.search.NASearcher instance",
        context="for the direct search phase",
    )
    def _call_searcher(searcher, parallel=True):
        searcher.run(parallel=parallel)
        return searcher.samples, searcher.objectives

    @error_handler(
        when="in creating neighpy.search.NASearcher object",
        context="at initialisation",
    )
    def _initialise_searcher(self):
        from neighpy import NASearcher

        return NASearcher(
            self.inv_problem.objective,
            ns=self._params["n_samples_per_iteration"],
            nr=self._params["n_cells_to_resample"],
            ni=self._params["n_initial_samples"],
            n=self._params["n_iterations"],
            bounds=self._params["bounds"],
        )



import pytest

from cofi import InversionOptions
from cofi.tools import BaseInferenceTool


def test_set_unset_solving_method(capsys):
    inv_options = InversionOptions()
    # 0
    inv_options.suggest_tools()
    console_output = capsys.readouterr()
    assert "optimization" in console_output.out
    assert "scipy.linalg.lstsq" in console_output.out
    inv_options.suggest_solving_methods()
    console_output = capsys.readouterr()
    assert "optimization" in console_output.out
    assert "`suggest_tools()` to see a full list of backend tools" in console_output.out
    # 1
    with pytest.raises(ValueError):
        inv_options.set_solving_method("abc")
    with pytest.raises(ValueError, match=".*Did you mean 'optimization'?.*"):
        inv_options.set_solving_method("optimizations")
    # 2
    inv_options.set_solving_method("matrix solvers")
    inv_options.suggest_tools()
    console_output = capsys.readouterr()
    assert "optimization" not in console_output.out
    assert "scipy.linalg.lstsq" in console_output.out
    # 3
    inv_options.unset_solving_method()
    inv_options.suggest_tools()
    console_output = capsys.readouterr()
    assert "optimization" in console_output.out
    assert "scipy.linalg.lstsq" in console_output.out
    # 4
    inv_options.set_solving_method("matrix solvers")
    inv_options.set_solving_method(None)
    inv_options.suggest_tools()
    console_output = capsys.readouterr()
    assert "optimization" in console_output.out
    assert "scipy.linalg.lstsq" in console_output.out


def test_set_unset_tool(capsys):
    inv_options = InversionOptions()
    # 0 - invalid input
    with pytest.raises(ValueError):
        inv_options.set_tool("abc")
    with pytest.raises(ValueError, match=".*Did you mean 'scipy.optimize.minimize'?.*"):
        inv_options.set_tool("scipy.minimize")
    # 1 - mismatch with solving_method
    inv_options.set_solving_method("optimization")
    with pytest.warns(UserWarning):
        inv_options.set_tool("scipy.linalg.lstsq")
    # 2 - unset
    inv_options.unset_tool()
    assert inv_options.tool is None
    # 3 - default without solving_method
    inv_options.unset_solving_method()
    assert inv_options.get_default_tool() == "scipy.optimize.minimize"
    inv_options.suggest_solver_params()
    console_output = capsys.readouterr()
    assert "default" in console_output.out
    assert "Required parameters" in console_output.out
    assert "Optional parameters & default settings" in console_output.out
    # 4 - default given solving_method
    inv_options.set_solving_method("optimization")
    assert inv_options.get_default_tool() == "scipy.optimize.minimize"

    # 5 - suggest solving method
    # 5 - set None
    inv_options.set_tool(inv_options.get_default_tool())
    inv_options.set_tool(None)
    assert inv_options.tool is None
    # 6 - self-defined tool
    class MyOwnSolver(BaseInferenceTool):
        def __init__(self, inv_problem, inv_options): pass
        def __call__(self): return 1

    inv_options.set_tool(MyOwnSolver)
    # 7 - self-defined invalid tool
    class AnotherSolver(BaseInferenceTool):
        def __init__(self, inv_problem, inv_options): pass

    with pytest.raises(ValueError):
        inv_options.set_tool(AnotherSolver)


def test_set_params():
    inv_options = InversionOptions()
    inv_options.set_params(a=3, b=5)
    assert inv_options.get_params()["a"] == 3
    assert inv_options.get_params()["b"] == 5


def test_summary(capsys):
    inv_options = InversionOptions()
    # 0
    inv_options.summary()
    console_output = capsys.readouterr()
    assert "Solving method: None set" in console_output.out
    assert "Solver-specific parameters: None set" in console_output.out
    assert "Backend tool" in console_output.out
    assert "(by default)" in console_output.out
    # 1
    inv_options.set_solving_method("optimization")
    inv_options.summary()
    console_output = capsys.readouterr()
    assert "Solving method: optimization" in console_output.out
    assert "Solver-specific parameters: None set" in console_output.out
    assert "Backend tool" in console_output.out
    assert "(by default)" in console_output.out
    # 2
    inv_options.set_tool(inv_options.get_default_tool())
    inv_options.summary()
    console_output = capsys.readouterr()
    assert "Solving method: optimization" in console_output.out
    assert "Solver-specific parameters: None set" in console_output.out
    assert "Backend tool" in console_output.out
    assert "(by default)" not in console_output.out
    # 3
    inv_options.set_params(a=3, b=5)
    inv_options.summary()
    console_output = capsys.readouterr()
    assert "Solving method: optimization" in console_output.out
    assert "Solver-specific parameters:" in console_output.out
    assert "Backend tool" in console_output.out
    assert "a = 3" in console_output.out
    assert "b = 5" in console_output.out


def test_repr():
    inv_options = InversionOptions()
    # 0
    rep = repr(inv_options)
    assert "InversionOptions" in rep
    assert "unknown" in rep
    assert "default" in rep
    # 1
    inv_options.set_solving_method("optimization")
    rep = repr(inv_options)
    assert "method='optimization'" in rep
    assert "unknown" not in rep
    # 2
    inv_options.set_tool(inv_options.get_default_tool())
    rep = repr(inv_options)
    assert "default" not in rep



import functools

from . import BaseInferenceTool, error_handler


class ScipyOptLstSq(BaseInferenceTool):
    documentation_links = [
        "https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html"
    ]
    short_description = (
        "SciPy's non-linear least squares solver with bounds on variables, "
        "algorithms include 'trf' (default), 'dogbox', and 'lm'"
    )

    @classmethod
    def required_in_problem(cls) -> set:
        return _init_class_methods()[0]

    @classmethod
    def optional_in_problem(cls) -> dict:
        return _init_class_methods()[1]

    @classmethod
    def required_in_options(cls) -> set:
        return _init_class_methods()[2]

    @classmethod
    def optional_in_options(cls) -> dict:
        return _init_class_methods()[3]

    @classmethod
    def available_algorithms(cls) -> set:
        return {"trf", "dogbox", "lm"}

    def __init__(self, inv_problem, inv_options):
        super().__init__(inv_problem, inv_options)
        self._components_used = list(self.required_in_problem())
        self._assign_args()

    def _assign_args(self):
        inv_problem = self.inv_problem
        # required_in_problem
        self._fun = inv_problem.residual
        self._x0 = inv_problem.initial_model
        # optional_in_problem
        defined_in_problem = self.inv_problem.defined_components()
        for component in self.optional_in_problem():
            if component in defined_in_problem:
                setattr(
                    self,
                    f"_{component}" if component != "jacobian" else "_jac",
                    getattr(self.inv_problem, component),
                )
                self._components_used.append(component)
            else:  # default
                setattr(
                    self,
                    f"_{component}" if component != "jacobian" else "_jac",
                    self.optional_in_problem()[component],
                )

    def __call__(self) -> dict:
        opt_result = self._call_np_least_squares()
        result = dict(opt_result.items())
        result["model"] = result.pop("x")
        return result

    @error_handler(
        when="when solving the least_squares optimization problem",
        context="calling `scipy.optimize.least_squares`",
    )
    def _call_np_least_squares(self):
        from scipy.optimize import least_squares

        return least_squares(
            fun=self._fun,
            x0=self._x0,
            jac=self._jac,
            bounds=self._bounds,
            method=self._params["method"],
            ftol=self._params["ftol"],
            xtol=self._params["xtol"],
            gtol=self._params["gtol"],
            x_scale=self._params["x_scale"],
            loss=self._params["loss"],
            f_scale=self._params["f_scale"],
            diff_step=self._params["diff_step"],
            tr_solver=self._params["tr_solver"],
            tr_options=self._params["tr_options"],
            jac_sparsity=self._params["jac_sparsity"],
            max_nfev=self._params["max_nfev"],
            verbose=self._params["verbose"],
            args=(),  # handled by cofi.BaseProblem
            kwargs={},  # handled by cofi.BaseProblem
        )


@functools.lru_cache(maxsize=None)
def _init_class_methods():
    """get a list of arguments and defaults for scipy.minimize.least_squares

    TODO arguments not supported by BaseProblem due to myself not sure how this can be
    handled for other backend tools: `args`, `kwargs`, `x_scale`, `loss`, `f_scale`
    """
    import inspect
    from scipy.optimize import least_squares

    _scipy_ls_args = dict(inspect.signature(least_squares).parameters)
    _scipy_ls_args["jacobian"] = _scipy_ls_args.pop("jac")
    required_in_problem = {"residual", "initial_model"}
    optional_in_problem = {
        k: v.default for k, v in _scipy_ls_args.items() if k in {"jacobian", "bounds"}
    }
    required_in_options = set()
    optional_in_options = {
        k: v.default
        for k, v in _scipy_ls_args.items()
        if k
        in {
            "method",
            "ftol",
            "xtol",
            "gtol",
            "x_scale",
            "f_scale",
            "loss",
            "diff_step",
            "tr_solver",
            "tr_options",
            "jac_sparsity",
            "max_nfev",
            "verbose",
        }
    }
    return (
        required_in_problem,
        optional_in_problem,
        required_in_options,
        optional_in_options,
    )


# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.least_squares -> trf
# description: Trust Region Reflective algorithm, particularly suitable for large sparse problems with bounds.

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.least_squares -> dogbox
# description: Dogleg algorithm with rectangular trust regions, typical use case is small problems with bounds.

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.least_squares -> lm
# description: Levenberg-Marquardt algorithm as implemented in MINPACK. Doesn’t handle bounds and sparse Jacobians.



from ._base_inference_tool import BaseInferenceTool, error_handler

from ._scipy_opt_min import ScipyOptMin
from ._scipy_opt_lstsq import ScipyOptLstSq
from ._scipy_lstsq import ScipyLstSq
from ._emcee import Emcee
from ._cofi_simple_newton import CoFISimpleNewton
from ._pytorch_optim import PyTorchOptim
from ._cofi_border_collie_optimization import CoFIBorderCollieOptimization
from ._bayes_bay import BayesBay
from ._neighpy import Neighpy, NeighpyI, NeighpyII
from ._scipy_sparse_lstsq import ScipySparseLstSq


__all__ = [
    "BaseInferenceTool",  # public API, for advanced usage (own solver)
    "ScipyOptMin",
    "ScipyOptLstSq",
    "ScipyLstSq",
    "Emcee",
    "CoFISimpleNewton",
    "PyTorchOptim",
    "BayesBay",
    "CoFIBorderCollieOptimization",
    "Neighpy",
    "NeighpyI",
    "NeighpyII",
    "ScipySparseLstSq",
]

# inference tools table grouped by method:
# {inv_options.method -> {inv_options.tool -> BaseInferenceTool}}
inference_tools_table = {
    "optimization": {
        "scipy.optimize.minimize": ScipyOptMin,
        "scipy.optimize.least_squares": ScipyOptLstSq,
        "torch.optim": PyTorchOptim,
        "cofi.border_collie_optimization": CoFIBorderCollieOptimization,
        "neighpyI": NeighpyI,
    },
    "matrix solvers": {
        "scipy.linalg.lstsq": ScipyLstSq,
        "cofi.simple_newton": CoFISimpleNewton,
        "scipy.sparse.linalg": ScipySparseLstSq,
    },
    "sampling": {
        "emcee": Emcee,
        "bayesbay": BayesBay,
        "neighpy": Neighpy,
        "neighpyII": NeighpyII,
    },
}

# inference tools suggest table grouped by method: {inv_options.method -> inv_options.tool}
# NOTE: the default backend solver is from this table, set the first one manually when necessary
# e.g. {'optimization': ['scipy.optimize.minimize'], 'matrix solvers': ['scipy.linalg.lstsq']}
tool_suggest_table = {k: list(val.keys()) for k, val in inference_tools_table.items()}

# tools dispatch table grouped by tool: {inv_options.tool -> BaseInferenceTool}
# e.g. {'scipy.optimize.minimize':
#           <class 'cofi.tools.scipy_opt_min.ScipyOptMinSolver'>,
#       'scipy.linalg.lstsq':
#           <class 'cofi.tools.scipy_lstsq.ScipyLstSqSolver'>}
tool_dispatch_table = {
    k: val for values in inference_tools_table.values() for k, val in values.items()
}

# all solving methods: {inv_options.method}
# e.g. {'optimization', 'matrix solvers'}
solving_methods = set(inference_tools_table.keys())

# alias for deprecated API
BaseSolver = BaseInferenceTool


# Extra information for InLab Explorer
# link_doc: CoFI -> https://cofi.readthedocs.io
# link_git: CoFI -> https://github.com/inlab-geo/cofi
# description: CoFI -> Common Framework for Inference
# description: Parameter estimation -> Parameter estimation is the process of determining the specific numerical values that define a parametrised (mathematical) model, often using methods like Maximum Likelihood Estimation or Least Squares to best fit the observed data.
# description: Ensemble methods -> Ensemble methods are a class of inference technique that result in multiple models, rather than a single model through parameter estimation, these might be driven by data fitting or Bayesian sampling.
# description: Optimization -> Optimization involves finding the best solution from a set of possible solutions, usually by minimizing or maximizing a certain function.
# description: Matrix based solvers -> Matrix-based solvers are computational algorithms that solve systems of equations, which often arise in linear or iteratively linear parameter estimation problems.
# description: Bayesian sampling -> Bayesian sampling is a technique for drawing samples (models) that follow a probability distribution of unknown parameters based on observed data and prior beliefs.
# description: Non linear -> Non-linear optimization focuses on finding the maximum, or minimum, of a function that is not necessarily quadratic over its parameters.
# description: Linear -> Optimization on linear problems involves finding the best solution from a set of possible solutions, where the function to be optimized is typically quadratic.
# description: Linear system solvers -> Linear system solvers are algorithms designed to find the values of unknowns in a set of linear equations.
# description: McMC samplers -> Markov chain Monte Carlo (McMC) samplers are algorithms for generating samples from complex probability distributions, often used in Bayesian inference.
# description: Direct search -> In the context of ensemble methods, direct search involves a search of the parameter space, e.g. to reduce a data misfit function, without requiring gradient information, These typically make use of exploration of the space and exploitation of previous sampling to guide search.
# description: Monte Carlo -> Monte Carlo methods use random sampling to obtain numerical results for problems that might be deterministic in principle.
# description: Deterministic -> Deterministic methods in direct search use specific rules, rather than randomness, to explore the parameter space for the best solution.
# description: Trans-D McMC -> Trans-Dimensional Markov chain Monte Carlo (Trans-D McMC) is a specialized form of McMC that allows for model selection by transitioning between different dimensional spaces.



import functools

from . import BaseInferenceTool, error_handler


class PyTorchOptim(BaseInferenceTool):
    documentation_links = [
        "https://pytorch.org/docs/stable/optim.html#algorithms",
    ]
    short_description = "PyTorch Optimizers under module `torch.optim`"

    @classmethod
    def required_in_problem(cls) -> set:
        return {"objective", "gradient", "initial_model"}

    @classmethod
    def optional_in_problem(cls) -> dict:
        return dict()

    @classmethod
    def required_in_options(cls) -> set:
        return {"algorithm", "num_iterations"}

    @classmethod
    def optional_in_options(cls) -> dict:
        return {
            "verbose": True,
            "callback": None,
            "algorithm_params": dict(),
        }

    @classmethod
    @functools.lru_cache(maxsize=None)
    def available_algorithms(cls) -> set:
        import torch

        optim_dir = dir(torch.optim)
        algs = {name for name in optim_dir if name[0].isupper() and name != "Optimizer"}
        return algs

    def __init__(self, inv_problem, inv_options):
        # save extra options into inv_options.hyper_params["algorithm_params"]
        if "algorithm_params" not in inv_options.hyper_params:
            inv_options.hyper_params["algorithm_params"] = dict()
        for param in list(inv_options.hyper_params):
            if (
                param not in self.optional_in_options()
                and param not in self.required_in_options()
            ):
                inv_options.hyper_params["algorithm_params"][param] = (
                    inv_options.hyper_params.pop(param)
                )

        # initialisation, validation
        super().__init__(inv_problem, inv_options)
        self._components_used = list(self.required_in_problem())
        self._validate_algorithm()

        # save problem info for later use
        self._obj = self.inv_problem.objective
        self._grad = self.inv_problem.gradient

        # initialize torch stuff
        self._initialize_torch_tensor()
        self._initialize_torch_optimizer()
        self._initialize_torch_objective()

        # initialize function evaluation counter
        self._nb_evaluations = 0

    def __call__(self) -> dict:
        import torch

        losses = []
        for i in range(self._params["num_iterations"]):
            self._one_iteration(i, losses)
        losses_out = torch.stack(losses)
        return {
            "model": self._m.detach().numpy(),
            "objective_value": self._last_loss.detach().numpy(),
            "losses": losses_out,
            "n_obj_evaluations": self._nb_evaluations,
            "n_grad_evaluations": self._nb_evaluations,
            "success": True,
        }

    def _validate_algorithm(self):
        if self._params["algorithm"] not in self.available_algorithms():
            raise ValueError(
                f"the algorithm you've chosen ({self._params['algorithm']}) "
                "is invalid. Please choose from the following: "
                f"{self.available_algorithms()}"
            )

    @error_handler(
        when="in converting initial_model into PyTorch Tensor with requires_grad=True",
        context="before solving the optimization problem",
    )
    def _initialize_torch_tensor(self):
        import torch

        if torch.is_tensor(self.inv_problem.initial_model):
            self._m = (
                self.inv_problem.initial_model.double()
                .clone()
                .detach()
                .requires_grad_(True)
            )
        else:
            self._m = torch.tensor(
                self.inv_problem.initial_model, dtype=float, requires_grad=True
            )

    @error_handler(
        when=f"in creating PyTorch Optimizer",
        context="before solving the optimization problem",
    )
    def _initialize_torch_optimizer(self):
        import torch

        self.torch_optimizer = getattr(torch.optim, self._params["algorithm"])(
            [self._m],
            **self._params["algorithm_params"],
        )

    @error_handler(
        when="in creating PyTorch custom Loss Function",
        context="before solving the optimization problem",
    )
    def _initialize_torch_objective(self):
        self.torch_objective = _CoFIObjective().apply

    @error_handler(
        when="when performing optimization stepping",
        context="in the process of solving",
    )
    def _one_iteration(self, i, losses):
        def closure():
            self.torch_optimizer.zero_grad()
            self._last_loss = self.torch_objective(self._m, self._obj, self._grad)
            self._last_loss.backward()
            self._nb_evaluations += 1
            return self._last_loss

        self.torch_optimizer.step(closure)
        losses.append(self._last_loss)
        if self._params["callback"] is not None:
            self._run_callback()
        if self._params["verbose"]:
            print(f"Iteration #{i}, objective value: {self._last_loss}")

    @error_handler(
        when="when running your callback function",
        context="in the process of solving",
    )
    def _run_callback(self):
        self._params["callback"](self._m)


def _CoFIObjective():
    import torch

    class CoFIObjective(torch.autograd.Function):
        # https://pytorch.org/docs/stable/generated/torch.autograd.Function.backward.html
        @staticmethod
        def forward(ctx, m, my_objective, my_gradient):
            # calculate and save gradient value
            grad = my_gradient(m)
            if not torch.is_tensor(grad):  # converting type only when not tensor
                grad = torch.tensor(grad)
            ctx.save_for_backward(grad)
            # calculate and return objective value
            obj_val = my_objective(m)
            if not torch.is_tensor(obj_val):  # converting type only when not tensor
                obj_val = torch.tensor(obj_val, requires_grad=True)
            return obj_val

        @staticmethod
        def backward(ctx, _):
            grad = ctx.saved_tensors[0]
            return grad, None, None

    return CoFIObjective


# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> Adadelta
# description: ADADELTA, an Adaptive Learning Rate Method.
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.Adadelta.html#torch.optim.Adadelta

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> Adagrad
# description: Adaptive Subgradient Methods for Online Learning and Stochastic Optimization.
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.Adagrad.html#torch.optim.Adagrad

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> Adam
# description: Adam, a Method for Stochastic Optimization.
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.Adam.html#torch.optim.Adam

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> AdamW
# description: Decoupled Weight Decay Regularization.
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.AdamW.html#torch.optim.AdamW

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> SparseAdam
# description: Lazy version of Adam algorithm suitable for sparse tensors.
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.SparseAdam.html#torch.optim.SparseAdam

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> Adamax
# description: Adamax algorithm (a variant of Adam based on infinity norm).
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.Adamax.html#torch.optim.Adamax

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> ASGD
# description: Averaged Stochastic Gradient Descent.
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.ASGD.html#torch.optim.ASGD

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> LBFGS
# description: L-BFGS algorithm, heavily inspired by minFunc.
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.LBFGS.html#torch.optim.LBFGS

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> NAdam
# description: NAdam algorithm, incorporating Nesterov Momentum into Adam
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.NAdam.html#torch.optim.NAdam

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> RAdam
# description: RAdam algorithm, on the variance of the adaptive learning rate and beyond
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.RAdam.html#torch.optim.RAdam

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> RMSprop
# description: RMSprop, a form of stochastic gradient descent where the gradients are divided by a running average of their recent magnitude
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.RMSprop.html#torch.optim.RMSprop

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> Rprop
# description: The resilient backpropagation algorithm.
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.Rprop.html#torch.optim.Rprop

# CoFI -> Parameter estimation -> Optimization -> Non linear -> torch.optim -> SGD
# description: The stochastic gradient descent (optionally with momentum).
# documentation: https://pytorch.org/docs/stable/generated/torch.optim.SGD.html#torch.optim.SGD



import os
import pytest
import numpy

from cofi import BaseProblem
import cofi.utils as utils
from cofi._exceptions import (
    InvalidOptionError, 
    InvocationError, 
    NotDefinedError, 
)


# ---------------- data ---------------------------------------------------------------
data_files_to_test = [
    "datasets/dummy_test1_comma.txt",
    "datasets/dummy_test2_tab.txt",
    "datasets/dummy_test3_idx.txt",
    "datasets/dummy_test4_npy.npy",
    "datasets/dummy_test5_pickle.pickle",
    "datasets/dummy_test6_pickle.pkl",
]


@pytest.fixture(params=data_files_to_test)
def data_path(request):
    path_to_current_file = os.path.realpath(__file__)
    current_directory = os.path.split(path_to_current_file)[0]
    data_path = os.path.join(current_directory, request.param)
    return data_path


def test_set_data_from_file(data_path):
    inv_problem = BaseProblem()
    if "idx" in data_path:
        inv_problem.set_data_from_file(data_path, 0)
    else:
        inv_problem.set_data_from_file(data_path)


def test_set_data_with_uncertainty():
    inv_problem = BaseProblem()
    inv_problem.set_data(numpy.ones((2,1)), numpy.zeros((2,2)), numpy.zeros((2,2)))
    assert inv_problem.data_defined
    assert inv_problem.data_covariance_defined
    assert inv_problem.data_covariance_inv_defined
    inv_problem.set_data_covariance(numpy.ones((2,2)))
    assert inv_problem.data_covariance_defined
    assert inv_problem.data_covariance[0,0] == 1


# ---------------- objective ----------------------------------------------------------
@pytest.fixture
def problem_objective_setup():
    inv_problem = BaseProblem()
    _x = numpy.array([1, 2, 3, 4, 5])
    _fwd = lambda m: m[0] + m[1] * _x + m[2] * _x**2
    _m_true = numpy.array([2, 1, 1])
    _y = _fwd(_m_true)
    _dt_misfit = lambda m: numpy.sum((_fwd(m) - _y)**2)
    _reg = lambda m: 0.5 * m.T @ m
    return inv_problem, _fwd, _y, _dt_misfit, _reg


def test_set_obj(problem_objective_setup):
    # setup instance
    inv_problem, _, _, _dt_misfit, _ = problem_objective_setup
    inv_problem.set_objective(_dt_misfit)
    # check
    assert inv_problem.objective_defined
    assert not inv_problem.gradient_defined
    assert not inv_problem.hessian_defined
    assert not inv_problem.residual_defined
    assert not inv_problem.jacobian_defined
    assert not inv_problem.data_misfit_defined
    assert not inv_problem.regularization_defined
    assert not inv_problem.forward_defined
    assert not inv_problem.data_defined
    assert len(inv_problem.defined_components()) == 1
    assert inv_problem.objective(numpy.array([2, 1, 1])) == 0
    assert inv_problem.objective(numpy.array([2, 1, 2])) == 979


# ---------------- data misfit --------------------------------------------------------
def test_set_data_misfit(problem_objective_setup, capsys):
    # setup instance
    inv_problem, _, _, _dt_misfit, _ = problem_objective_setup
    inv_problem.set_data_misfit(_dt_misfit)
    # check
    assert inv_problem.data_misfit_defined
    assert inv_problem.objective_defined
    assert len(inv_problem.defined_components()) == 2
    assert inv_problem.objective(numpy.array([2, 1, 1])) == 0
    inv_problem.summary()
    captured = capsys.readouterr()
    assert "you did not set regularization" in captured.out


# ---------------- regularization -----------------------------------------------------
def test_set_data_misfit_regularization(problem_objective_setup):
    # setup instance
    inv_problem, _, _, _dt_misfit, _reg = problem_objective_setup
    inv_problem.set_data_misfit(_dt_misfit)
    inv_problem.set_regularization(_reg)
    # check
    assert inv_problem.data_misfit_defined
    assert inv_problem.regularization_defined
    assert inv_problem.objective_defined
    assert len(inv_problem.defined_components()) == 3
    assert inv_problem.objective(numpy.array([2, 1, 1])) == 3
    assert inv_problem.objective(numpy.array([2, 1, 2])) == 983.5


def test_set_regularization_from_utils_quadratic_reg(problem_objective_setup):
    # setup instance
    inv_problem, _, _, _dt_misfit, _ = problem_objective_setup
    inv_problem.set_data_misfit(_dt_misfit)
    _my_reg_from_utils = utils.QuadraticReg(model_shape=(3,))
    inv_problem.set_regularization(_my_reg_from_utils)
    # check
    assert len(inv_problem.defined_components()) == 4
    assert inv_problem.regularization(numpy.array([2, 1, 1])) == 6
    assert inv_problem.regularization(numpy.array([2, 1, 2])) == 9


def test_set_regularization_from_utils_gaussian_prior(problem_objective_setup):
    # setup instance
    inv_problem, _, _, _dt_misfit, _ = problem_objective_setup
    inv_problem.set_data_misfit(_dt_misfit)
    _my_reg_from_utils = utils.GaussianPrior(((3,), 0.5), numpy.ones(3))
    inv_problem.set_regularization(_my_reg_from_utils)
    # check
    assert len(inv_problem.defined_components()) == 3
    assert numpy.isclose(inv_problem.regularization(numpy.array([2, 1, 1])), 8.2205933)
    assert numpy.isclose(inv_problem.regularization(numpy.array([2, 1, 2])), 16.441187)

def test_set_regularization_both_none():
    inv_problem = BaseProblem()
    with pytest.raises(ValueError, match=r".*`regularization` or `regularization_matrix`.*"):
        inv_problem.set_regularization(None, None)

# ---------------- forward + data -----------------------------------------------------
def test_set_forward_data(problem_objective_setup):
    # setup instance
    inv_problem, _fwd, _y, _, _ = problem_objective_setup
    inv_problem.set_forward(_fwd)
    inv_problem.set_data(_y)
    inv_problem.set_data_misfit("squared error")
    # check
    assert len(inv_problem.defined_components()) == 5
    assert inv_problem.data_misfit(numpy.array([2, 1, 1])) == 0


def test_set_forward_data_with_Cd(problem_objective_setup):
    # setup instance
    inv_problem, _fwd, _y, _, _ = problem_objective_setup
    inv_problem.set_forward(_fwd)
    inv_problem.set_data(_y)
    inv_problem.set_data_misfit("squared error")
    inv_problem.set_data_covariance(0.5 * numpy.eye(5))
    # check
    assert len(inv_problem.defined_components()) == 6
    assert inv_problem.data_misfit(numpy.array([2, 1, 1])) == 0
    assert inv_problem.data_misfit(numpy.array([2, 1, 2])) == 979 * 2


def test_set_forward_data_with_Cdinv(problem_objective_setup):
    # setup instance
    inv_problem, _fwd, _y, _, _ = problem_objective_setup
    inv_problem.set_forward(_fwd)
    inv_problem.set_data(_y)
    inv_problem.set_data_misfit("squared error")
    inv_problem.set_data_covariance_inv(0.5 * numpy.eye(5))
    # check
    assert len(inv_problem.defined_components()) == 6
    assert inv_problem.data_misfit(numpy.array([2, 1, 1])) == 0
    assert inv_problem.data_misfit(numpy.array([2, 1, 2])) == 979 / 2

def test_set_forward_data_with_Cdinv_non_diagonal(problem_objective_setup):
    # setup instance
    inv_problem, _fwd, _y, _, _ = problem_objective_setup
    inv_problem.set_forward(_fwd)
    inv_problem.set_data(_y)
    inv_problem.set_data_misfit("squared error")
    Cdinv = 0.5 * numpy.eye(5)
    Cdinv[0,1] = 1
    inv_problem.set_data_covariance_inv(Cdinv)
    # check
    assert len(inv_problem.defined_components()) == 6
    assert inv_problem.data_misfit(numpy.array([2, 1, 1])) == 0
    assert inv_problem.data_misfit(numpy.array([2, 1, 2])) == 493.5

def test_set_invalid_misfit_options():
    inv_problem = BaseProblem()
    with pytest.raises(InvalidOptionError):
        inv_problem.set_data_misfit("FOO")
    inv_problem.set_data_misfit("squared error")
    with pytest.raises(InvocationError):
        inv_problem.data_misfit(numpy.array([1, 2, 3]))


# ---------------- model covariance as extra information ------------------------------
def test_model_cov_from_data_cov():
    inv_problem = BaseProblem()
    sigma = 1.0
    Cdinv = numpy.eye(100)/(sigma**2)
    inv_problem.set_data_covariance_inv(Cdinv)
    with pytest.raises(NotDefinedError, match=r".*`jacobian` is called.*"):
        inv_problem.model_covariance_inv(None)
    inv_problem.set_jacobian(numpy.array([[n**i for i in range(2)] for n in range(100)]))
    inv_problem.model_covariance(None)


# ---------------- jac/hess times vector (auto generated) -----------------------------
def test_hess_times_vector():
    inv_problem = BaseProblem()
    # 1
    _hess = numpy.array([[1,0],[0,2]])
    inv_problem.set_hessian(_hess)
    _test_res = inv_problem.hessian_times_vector(0, numpy.array([1,1]))
    assert numpy.array_equal(_test_res, numpy.array([1,2]))
    # 2
    inv_problem.set_hessian(lambda _: _hess)
    _test_res = inv_problem.hessian_times_vector(0, numpy.array([1,1]))
    assert numpy.array_equal(_test_res, numpy.array([1,2]))



import sys
from typing import Any, List, Tuple, Union


GITHUB_ISSUE = "https://github.com/inlab-geo/cofi/issues"


class CofiError(Exception):
    """Base class for all CoFI errors"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _form_str(self, super_msg, msg):
        return f"{msg}\n\n{super_msg}" if super_msg else msg


class InvalidOptionError(CofiError, ValueError):
    r"""Raised when user passes an invalid option into our methods / functions

    This is a subclass of :exc:`CofiError` and :exc:`ValueError`.

    Parameters
    ----------
    *args : Any
        passed on directly to :exc:`ValueError`
    name: str
        name of the item that tries to take the invalid option
    invalid_option : Any
        the invalid option entered
    valid_options : list or str
        a list of valid options to choose from, or a string describing valid options
    """

    def __init__(
        self,
        *args,
        name: str,
        invalid_option: Any,
        valid_options: Union[List, str],
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._name = name
        self._invalid_option = invalid_option
        self._valid_options = valid_options

    def __str__(self) -> str:
        super_msg = super().__str__()
        msg = (
            f"the {self._name} you've entered ('{self._invalid_option}') is "
            f"invalid, please choose from the following: {self._valid_options}.\n\n"
            f"If you find it valuable to have '{self._invalid_option}' for "
            f"{self._name} in CoFI, please create an issue here: {GITHUB_ISSUE}"
        )
        return self._form_str(super_msg, msg)


class DimensionMismatchError(CofiError, ValueError):
    r"""Raised when model or data shape doesn't match existing problem settings

    This is a subclass of :exc:`CofiError` and :exc:`ValueError`.

    Parameters
    ----------
    *args : Any
        passed on directly to :exc:`ValueError`
    entered_dimension : tuple
        dimension entered that conflicts with existing one
    entered_name : str
        name of the item, the dimension of which is entered
    expected_dimension : tuple
        dimension expected based on existing information
    expected_source : str
        name of an existing component that infers ``expected_dimension``
    """

    def __init__(
        self,
        *args,
        entered_dimension: Tuple,
        entered_name: str,
        expected_dimension: Tuple,
        expected_source: str,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._entered_dimension = entered_dimension
        self._entered_name = entered_name
        self._expected_dimension = expected_dimension
        self._expected_source = expected_source

    def __str__(self) -> str:
        super_msg = super().__str__()
        msg = (
            f"the {self._entered_name} you've provided (shape: "
            f"{self._entered_dimension}) doesn't match and cannot be reshaped "
            f"into the dimension you've set for {self._expected_source} which is "
            f"{self._expected_dimension}"
        )
        return self._form_str(super_msg, msg)


class NotDefinedError(CofiError, NotImplementedError):
    r"""Raised when a certain property or function is not set to a :class:BaseProblem
    instance but attempts are made to use it (e.g. in a solving approach)

    This is a subclass of :exc:`CofiError` and :exc:`NotImplementedError`.

    Parameters
    ----------
    *args : Any
        passed on directly to :exc:`NotImplementedError`
    needs : list or str
        a list of information required to perform the operation, or a string describing
        them
    """

    def __init__(self, *args, needs: Union[List, str], **kwargs):
        super().__init__(*args, **kwargs)
        self._needs = needs

    def __str__(self) -> str:
        super_msg = super().__str__()
        msg = (
            f"`{self._needs}` is called but you haven't implemented or added it to the"
            " problem setup"
        )
        return self._form_str(super_msg, msg)


class InvocationError(CofiError, RuntimeError):
    r"""Raised when there's an error happening during excecution of a function

    This is a subclass of :exc:`CofiError` and :exc:`RuntimeError`.

    One should raise this error by ``raise InvocationError(func_name=a, autogen=b) from exception``,
    where ``exception`` is the original exception caught

    Parameters
    ----------
    *args : Any
        passed on directly to :exc:`RuntimeError`
    func_name : str
        name of the function that runs into error
    autogen : bool
        whether this function is automatically generated or defined by users
    """

    def __init__(self, *args, func_name: str, autogen: bool, **kwargs):
        super().__init__(*args, **kwargs)
        self._func_name = func_name
        self._func_name_prefix = "auto-generated" if autogen else "your"

    def __str__(self) -> str:
        super_msg = super().__str__()
        msg = (
            f"exception while calling {self._func_name_prefix} {self._func_name}. "
            "Check exception details from message above. "
        )
        return self._form_str(super_msg, msg)



import pytest

def test_deprecated_cofi_solvers():
    from cofi import solvers
    import cofi.solvers
    from cofi.solvers import BaseInferenceTool

def test_deprecated_base_solver():
    from cofi.solvers import BaseSolver
    with pytest.raises(TypeError):
        a = BaseSolver()



from . import BaseInferenceTool, error_handler


class NeighpyII(BaseInferenceTool):
    r"""Wrapper for the tool :code:`neighpy`, a Python implementation of the Neighbourhood Algorithm.

    The Neighbourhood Algorithm is a direct search method that uses a set of points in the parameter space to explore the objective function. It is a derivative-free method that is particularly useful for high-dimensional problems.

    This wrapper only runs the second phase of the NA i.e. the appraisal phase.
    This phase is used to refine the samples obtained from the first phase.
    It thus requires an initial set of samples with their corresponding log posteriors to be provided.
    This is initial set can be obtained either from the first phase of the NA or any other sampling method.
    """

    documentation_links = ["https://neighpy.readthedocs.io/en/latest/index.html"]
    short_description = [
        "Wrapper for Neighpy, a Python implementation of the Neighbourhood Algorithm"
    ]

    @classmethod
    def required_in_problem(cls) -> set:
        return set()

    @classmethod
    def optional_in_problem(cls) -> dict:
        return {}

    @classmethod
    def required_in_options(cls) -> set:
        return {
            "bounds",
            "initial_ensemble",
            "log_ppd",
            "n_resample",
            "n_walkers",
        }

    @classmethod
    def optional_in_options(cls) -> dict:
        return {}

    def __init__(self, inv_problem, inv_options):
        super().__init__(inv_problem, inv_options)
        self._params["ndim"] = len(self._params["bounds"])
        if (
            self._params["initial_ensemble"].shape[0]
            != self._params["log_ppd"].shape[0]
        ):
            raise ValueError(
                "The number of samples in initial_ensemble does not match the number of samples in log_ppd"
            )
        if self._params["initial_ensemble"].shape[1] != self._params["ndim"]:
            raise ValueError(
                "The shape of initial_ensemble does not match the number of dimensions"
            )
        self._components_used = list(self.required_in_problem())

    def __call__(self) -> dict:
        raw_results = self._call_backend_tool()
        res = {"success": True}
        return {**res, **raw_results}

    @error_handler(
        when="at some point in the Neighbourhood Algorithm Appraisal phase",
        context="",
    )
    def _call_backend_tool(self):

        appraiser = self._initalise_appraiser()
        appraisal_samples = self._call_appriaser(appraiser)

        return {"new_samples": appraisal_samples}

    @staticmethod
    @error_handler(
        when="in calling neighpy.appraise.NAAppraiser instance",
        context="for the appraisal phase",
    )
    def _call_appriaser(appraiser):
        appraiser.run()
        return appraiser.samples

    @error_handler(
        when="in creating neighpy.appraise.NAAppraiser object",
        context="at initialisation.",
    )
    def _initalise_appraiser(self):
        from neighpy import NAAppraiser

        return NAAppraiser(
            initial_ensemble=self._params["initial_ensemble"],
            log_ppd=self._params["log_ppd"],
            n_resample=self._params["n_resample"],
            n_walkers=self._params["n_walkers"],
            bounds=self._params["bounds"],
        )



import numpy as np
import copy

from . import BaseInferenceTool, error_handler


class CoFIBorderCollieOptimization(BaseInferenceTool):
    r"""Implentation of a Border Collie Optimization Algorithm

    Based on the concepts and equations given in T. Dutta, S. Bhattacharyya, S. Dey and
    J. Platos, "Border Collie Optimization," in IEEE Access, vol. 8, pp. 109177-109197,
    2020, doi: 10.1109/ACCESS.2020.2999540 https://ieeexplore.ieee.org/document/9106341
    """

    documentation_links = []
    short_description = "CoFI's implemntation of Border Collie Optimization"

    @classmethod
    def required_in_problem(cls) -> set:
        return {"objective", "bounds", "model_shape"}

    @classmethod
    def optional_in_problem(cls) -> dict:
        return {"initial_model": []}

    @classmethod
    def required_in_options(cls) -> set:
        return set()

    @classmethod
    def optional_in_options(cls) -> dict:
        return {"number_of_iterations": 50, "flock_size": 50, "seed": 42}

    ### inner classes for dog and sheep

    class dog:
        def __init__(self, dim):
            self.pos = np.zeros(dim)
            self.vel = np.zeros(dim)
            self.acc = np.zeros(dim)
            self.tim = np.zeros(dim)
            self.fit = 0
            pass

    class sheep:
        def __init__(self, dim):
            self.pos = np.zeros(dim)
            self.vel = np.zeros(dim)
            self.acc = np.zeros(dim)
            self.tim = np.zeros(dim)
            self.eyed = 0
            self.fit = 0
            pass

    def __init__(self, inv_problem, inv_options):
        super().__init__(inv_problem, inv_options)
        self._components_used = list(self.required_in_problem())
        np.seterr(divide="ignore", invalid="ignore")
        # self model size
        self.mod_size = np.prod(inv_problem.model_shape)
        # lower and upper boundaries
        self.lower_bounds = []
        self.upper_bounds = []
        for i in range(self.mod_size):
            self.lower_bounds.append(inv_problem.bounds[i][0])
            self.upper_bounds.append(inv_problem.bounds[i][1])

        self.lower_bounds = np.array(self.lower_bounds)
        self.upper_bounds = np.array(self.upper_bounds)

        try:
            self.initial_model = inv_problem.initial_model
        except:
            self.initial_model = self.optional_in_problem()["initial_model"]
        self.rng = np.random.default_rng(self._params["seed"])
        # number of iterations
        self.number_of_iterations = self._params["number_of_iterations"]

    def __call__(self) -> dict:
        self.pack_size = 3
        self.pack = []
        self.flock_size = self._params["flock_size"]
        self.flock = []
        self.initialise()
        for itr in range(self.number_of_iterations):
            self.update_fitness()
            # print( self.itr_min_pos,"|",self.itr_min_fit)
            # determin if a sheep needs eyeing
            if itr > 0:
                for idx, sheep in enumerate(self.flock):
                    if self.flock_fit[idx] > self.flock_fit_prev[idx]:
                        sheep.eyed += 1

            if itr == 0 or fit_min > self.itr_min_fit:
                fit_min = self.itr_min_fit

                res = {
                    "success": "maybe",
                    "model": self.itr_min_pos,
                    "pack_fitness_history": self.pack_fit_hist,
                    "flock_fitness_history": self.flock_fit_hist,
                    "pack_position_history": self.pack_pos_hist,
                    "flock_position_history": self.flock_pos_hist,
                }

                dpos = []
                for dog in self.pack:
                    dpos.append(dog.pos)
                self.pack_pos_hist.append(dpos)
                spos = []

                for sheep in self.flock:
                    spos.append(sheep.pos)
                self.flock_pos_hist.append(spos)

                self.dog2sheep2dog()

                dpos = []
                for dog in self.pack:
                    dpos.append(dog.pos)
                self.pack_pos_hist.append(dpos)
                spos = []

                for sheep in self.flock:
                    spos.append(sheep.pos)
                self.flock_pos_hist.append(spos)

            else:
                self.dog2sheep2dog()

            self.update_movement()
            self.check_positions()

        return res

    # The following are useful functions used by the main optimiser but it makes sense
    # to keep them seperate

    def initialise(self):
        for i in range(self.pack_size):
            self.pack.append(self.dog(self.mod_size))
            if isinstance(self.initial_model, np.ndarray) or self.initial_model:
                self.pack[i].pos = self.initial_model
            else:
                self.pack[i].pos = (
                    self.rng.uniform(size=self.mod_size)
                    * (self.upper_bounds - self.lower_bounds)
                    + self.lower_bounds
                )
            self.pack[i].vel = self.rng.uniform(size=self.mod_size)
            self.pack[i].acc = self.rng.uniform(size=self.mod_size)
            self.pack[i].tim = self.rng.uniform(size=self.mod_size)
        for i in range(self.flock_size):
            self.flock.append(self.sheep(self.mod_size))
            if isinstance(self.initial_model, np.ndarray) or self.initial_model:
                self.flock[i].pos = self.initial_model
            else:
                self.flock[i].pos = (
                    np.random.rand(self.mod_size)
                    * (self.upper_bounds - self.lower_bounds)
                    + self.lower_bounds
                )
            self.flock[i].vel = self.rng.uniform(size=self.mod_size)
            self.flock[i].acc = self.rng.uniform(size=self.mod_size)
            self.flock[i].tim = self.rng.uniform(size=self.mod_size)
            # print(self.flock[i].pos)

        self.pack_fit = np.zeros(self.pack_size)
        self.flock_fit = np.zeros(self.flock_size)
        self.pop_fit = np.zeros(self.pack_size + self.flock_size)

        self.pack_fit_hist = []
        self.flock_fit_hist = []

        self.pack_pos_hist = []
        self.flock_pos_hist = []

        return

    def update_fitness(self):
        # could be done parallel
        self.flock_fit_prev = copy.copy(self.flock_fit)

        for i, dog in enumerate(self.pack):
            self.pack_fit[i] = self.inv_problem.objective(self.pack[i].pos)
            # print(i,self.pack_fit[i],self.pack[i].pos)
            self.pack[i].fit = self.pack_fit[i]
            self.pop_fit[i] = self.pack_fit[i]
        for i, sheep in enumerate(self.flock):
            self.flock_fit[i] = self.inv_problem.objective(self.flock[i].pos)
            self.flock[i].fit = self.flock_fit[i]
            self.pop_fit[i + self.pack_size] = self.flock_fit[i]

        self.pack_fit_min_idx = np.argmin(self.pack_fit)
        self.pack_fit_min_val = self.pack_fit[self.pack_fit_min_idx]

        self.flock_fit_min_idx = np.argmin(self.flock_fit)
        self.flock_fit_min_val = self.flock_fit[self.flock_fit_min_idx]

        self.pack_fit_hist.append(self.pack_fit)

        if self.flock_fit_min_val < self.pack_fit_min_val:
            self.itr_min_pos = self.flock[self.flock_fit_min_idx].pos
            self.itr_min_fit = self.flock_fit_min_val
        else:
            self.itr_min_pos = self.pack[self.pack_fit_min_idx].pos
            self.itr_min_fit = self.pack_fit_min_val

        return

    def dog2sheep2dog(self):
        idx = self.pop_fit.argsort()
        popl = copy.copy(self.pack + self.flock)

        for i, dog in enumerate(self.pack):
            self.pack[i].pos = copy.copy(popl[idx[i]].pos)
            self.pack[i].vel = copy.copy(popl[idx[i]].vel)
            self.pack[i].acc = copy.copy(popl[idx[i]].acc)
            self.pack[i].tim = copy.copy(popl[idx[i]].tim)
            self.pack[i].fit = copy.copy(popl[idx[i]].fit)

        for i, sheep in enumerate(self.flock):
            self.flock[i].pos = copy.copy(popl[idx[i + self.pack_size]].pos)
            self.flock[i].vel = copy.copy(popl[idx[i + self.pack_size]].vel)
            self.flock[i].acc = copy.copy(popl[idx[i + self.pack_size]].acc)
            self.flock[i].tim = copy.copy(popl[idx[i + self.pack_size]].tim)
            self.flock[i].fit = copy.copy(popl[idx[i + self.pack_size]].fit)
        return

    def update_movement(self):
        # tournament selection for choosing left and right dogs
        if np.random.randint(1, high=3) == 1:
            dog = copy.copy(self.pack[1])
            self.pack[2] = copy.copy(self.pack[1])
            self.pack[1] = copy.copy(dog)

        # update dog velocity, acceleration and time
        for i, dog in enumerate(self.pack):
            v0 = self.pack[i].vel
            ## self.pack[i].vel=self.pack[i].vel+self.pack[i].acc
            try:
                self.pack[i].vel = np.sqrt(
                    self.pack[i].vel ** 2 + 2.0 * self.pack[i].acc * self.pack[i].pos
                )
                assert not np.any(np.isnan(self.pack[i].vel))
            except:
                self.pack[i].vel = np.zeros(np.shape(self.pack[i].pos))
            self.pack[i].acc = (self.flock[i].vel - v0) / self.pack[i].tim
            self.pack[i].tim = np.mean((self.pack[i].vel - v0) / self.pack[i].tim)

        # update dog position
        for i, dog in enumerate(self.pack):
            ## self.pack[i].pos=self.pack[i].pos+self.pack[i].vel
            self.pack[i].pos = (
                self.pack[i].vel * self.pack[i].tim
                + 0.5 * self.pack[i].acc * self.pack[i].tim ** 2
            )

        # update sheep velocity acceleration and time
        for i, sheep in enumerate(self.flock):
            dg = (self.pack[0].fit - self.flock[i].fit) - (
                ((self.pack[1].fit + self.pack[2].fit) / 2.0) - self.flock[i].fit
            )
            v0 = self.flock[i].vel
            if dg >= 0.0:  # sheep is gathered
                ## self.flock[i].vel=self.pack[0].vel+self.pack[0].acc/self.pack[0].tim
                try:
                    self.flock[i].vel = np.sqrt(
                        self.flock[i].vel + 2.0 * self.flock[i].acc * self.flock[i].pos
                    )
                    assert not np.any(np.isnan(self.flock[i].vel))
                except:
                    self.flock[i].vel = np.zeros(np.shape(self.flock[i].pos))

            if dg < 0.0:
                if self.flock[i].eyed < 5:
                    tan_theta = np.tan(self.rng.integers(1, high=90))
                    ##vl = self.pack[1].vel*tan_theta+self.pack[1].acc
                    vl = np.sqrt(
                        (self.pack[1].vel * tan_theta) ** 2
                        + 2.0 * self.pack[1].acc * self.pack[1].pos
                    )
                    tan_theta = np.tan(self.rng.integers(91, high=180))
                    ##vr = self.pack[2].vel*tan_theta+self.pack[2].acc
                    vr = np.sqrt(
                        (self.pack[2].vel * tan_theta) ** 2
                        + 2.0 * self.pack[2].acc * self.pack[2].pos
                    )
                    vs = (vl + vr) / 2.0
                    self.flock[i].vel = vs

                else:
                    self.flock[i].eyed = 0
                    if self.pack[1].fit < self.pack[2].fit:
                        ## self.flock[i].vel=self.pack[2].vel-self.pack[2].acc/self.pack[2].tim
                        self.flock[i].vel = np.sqrt(
                            self.pack[2].vel ** 2
                            - 2.0 * self.pack[2].acc * self.pack[2].pos
                        )
                    else:
                        ## self.flock[i].vel=self.pack[1].vel-self.pack[1].acc/self.pack[1].tim
                        self.flock[i].vel = np.sqrt(
                            self.pack[1].vel ** 2
                            - 2.0 * self.pack[1].acc * self.pack[1].pos
                        )
            self.flock[i].acc = (self.flock[i].vel - v0) / self.flock[i].tim
            self.flock[i].tim = np.mean((self.flock[i].vel - v0) / self.flock[i].acc)

            if dg >= 0.0:
                self.flock[i].pos = (
                    self.flock[i].vel * self.flock[i].tim
                    + 0.5 * self.flock[i].acc * self.flock[i].tim ** 2
                )
            if dg < 0.0:
                self.flock[i].pos = (
                    self.flock[i].vel * self.flock[i].tim
                    - 0.5 * self.flock[i].acc * self.flock[i].tim ** 2
                )
        return

    def check_positions(self):
        for i in range(self.pack_size):
            for j in range(self.mod_size):
                if (
                    self.pack[i].pos[j] >= self.upper_bounds[j]
                    or self.pack[i].pos[j] <= self.lower_bounds[j]
                ):
                    self.pack[i].pos[j] = (
                        self.rng.uniform()
                        * (self.upper_bounds[j] - self.lower_bounds[j])
                        + self.lower_bounds[j]
                    )
                    self.pack[i].acc[j] = self.rng.uniform()
                    self.pack[i].tim = self.rng.uniform()
                if np.isnan(self.pack[i].acc[j]) or self.pack[i].acc[j] == 0:
                    self.pack[i].pos[j] = (
                        self.rng.uniform()
                        * (self.upper_bounds[j] - self.lower_bounds[j])
                        + self.lower_bounds[j]
                    )
                    self.pack[i].acc[j] = self.rng.uniform()
                    self.pack[i].tim = self.rng.uniform()
                if np.isnan(self.pack[i].vel[j]) or self.pack[i].vel[j] == 0:
                    self.pack[i].pos[j] = (
                        self.rng.uniform()
                        * (self.upper_bounds[j] - self.lower_bounds[j])
                        + self.lower_bounds[j]
                    )
                    self.pack[i].acc[j] = self.rng.uniform()
                    self.pack[i].tim = self.rng.uniform()
                if np.isnan(self.pack[i].tim) or self.pack[i].tim == 0:
                    self.pack[i].pos[j] = (
                        self.rng.uniform()
                        * (self.upper_bounds[j] - self.lower_bounds[j])
                        + self.lower_bounds[j]
                    )
                    self.pack[i].acc[j] = self.rng.uniform()
                    self.pack[i].tim = self.rng.uniform()

        for i in range(self.flock_size):
            for j in range(self.mod_size):
                if (
                    self.flock[i].pos[j] >= self.upper_bounds[j]
                    or self.flock[i].pos[j] <= self.lower_bounds[j]
                ):
                    self.flock[i].pos[j] = (
                        self.rng.uniform()
                        * (self.upper_bounds[j] - self.lower_bounds[j])
                        + self.lower_bounds[j]
                    )
                    self.flock[i].acc[j] = self.rng.uniform()
                    self.flock[i].tim = self.rng.uniform()
                if np.isnan(self.flock[i].acc[j]) or self.flock[i].acc[j] == 0:
                    self.flock[i].pos[j] = (
                        self.rng.uniform()
                        * (self.upper_bounds[j] - self.lower_bounds[j])
                        + self.lower_bounds[j]
                    )
                    self.flock[i].acc[j] = self.rng.uniform()
                    self.flock[i].tim = self.rng.uniform()
                if np.isnan(self.flock[i].vel[j]) or self.flock[i].vel[j] == 0:
                    self.flock[i].pos[j] = (
                        self.rng.uniform()
                        * (self.upper_bounds[j] - self.lower_bounds[j])
                        + self.lower_bounds[j]
                    )
                    self.flock[i].acc[j] = self.rng.uniform()
                    self.flock[i].tim = self.rng.uniform()
                if np.isnan(self.flock[i].tim) or self.flock[i].tim == 0:
                    self.flock[i].pos[j] = (
                        self.rng.uniform()
                        * (self.upper_bounds[j] - self.lower_bounds[j])
                        + self.lower_bounds[j]
                    )
                    self.flock[i].acc[j] = self.rng.uniform()
                    self.flock[i].tim = self.rng.uniform()
        return


# CoFI -> Ensemble methods -> Direct search -> Monte Carlo -> cofi.border_collie_optimization -> Border Collie Optimization Algorithm
# description: Implementation of a Border Collie Optimization Algorithm, based on the concepts and equations given in Dutta et al., IEEE Access, 2020.
# documentation: https://ieeexplore.ieee.org/document/9106341



from abc import abstractmethod, ABCMeta
from numbers import Number
from functools import reduce
import numpy as np

from .._exceptions import DimensionMismatchError


class BaseRegularization(metaclass=ABCMeta):
    r"""Base class for a regularization term

    Check :class:`QuadraticReg` for a concrete example.

    .. rubric:: Basic interface

    The basic properties / methods for a regularization term in ``cofi.utils``
    include the following:

    .. autosummary::
        BaseRegularization.model_size
        BaseRegularization.reg
        BaseRegularization.gradient
        BaseRegularization.hessian
        BaseRegularization.__call__

    .. rubric:: Adding two terms

    Two instances of ``BaseRegularization`` can also be added together using the ``+``
    operator:

    .. autosummary::
        BaseRegularization.__add__

    """

    def __init__(
        self,
    ):
        pass

    @property
    @abstractmethod
    def model_shape(self) -> tuple:
        """the shape of models that current regularization function accepts"""
        raise NotImplementedError

    @property
    def model_size(self) -> Number:
        """the number of unknowns that current regularization function accepts"""
        return reduce(lambda a, b: a * b, np.array(self.model_shape), 1)

    def __call__(self, model: np.ndarray) -> Number:
        r"""a class instance itself can also be called as a function

        It works exactly the same as :meth:`reg`.

        In other words, the following two usages are exactly the same::

        >>> my_reg = QuadraticReg(factor=1, model_size=3)
        >>> my_reg_value = my_reg(np.array([1,2,3]))            # usage 1
        >>> my_reg_value = my_reg.reg(np.array([1,2,3]))        # usage 2
        """
        return self.reg(model)

    @abstractmethod
    def reg(self, model: np.ndarray) -> Number:
        """the regularization function value given a model to evaluate"""
        raise NotImplementedError

    @abstractmethod
    def gradient(self, model: np.ndarray) -> np.ndarray:
        """the gradient of regularization function with respect to model given a model

        The usual size for the gradient is :math:`(M,)` where :math:`M` is the number
        of model parameters
        """
        raise NotImplementedError

    @abstractmethod
    def hessian(self, model: np.ndarray) -> np.ndarray:
        """the hessian of regularization function with respect to model given a model

        The usual size for the Hessian is :math:`(M,M)` where :math:`M` is the number
        of model parameters
        """
        raise NotImplementedError

    def __add__(self, other_reg):
        r"""Adds two regularization terms

        Parameters
        ----------
        other_reg : BaseRegularization
            the second argument of "+" operator; must also be a
            :class:`BaseRegularization` instance

        Returns
        -------
        BaseRegularization
            a regularization term ``resRegularization`` such that:

            - :math:`\text{resRegularization.reg}(m)=\text{self.reg}(m)+\text{other_reg.reg}(m)`
            - :math:`\text{resRegularization.gradient}(m)=\text{self.gradient}(m)+\text{other_reg.gradient}(m)`
            - :math:`\text{resRegularization.hessian}(m)=\text{self.hessian}(m)+\text{other_reg.hessian}(m)`

        Raises
        ------
        TypeError
            when the ``other_reg`` is not a regularization term generated by CoFI Utils
        DimensionMismatchError
            when the ``other_reg`` doesn't accept model_size that matches the one of
            ``self``

        Examples
        --------

        >>> from cofi import BaseProblem
        >>> from cofi.utils import QuadraticReg
        >>> reg1 = QuadraticReg(model_shape=(3,), weighting_matrix="damping")
        >>> reg2 = QuadraticReg(model_shape=(3,), weighting_matrix="smoothing")
        >>> my_problem = BaseProblem()
        >>> my_problem.set_regularization(reg1 + reg2)

        """
        if not isinstance(other_reg, BaseRegularization):
            raise TypeError(
                f"unsupported operand type(s) for +: '{self.__class__.__name__}' "
                f"and '{other_reg.__class__.__name__}"
            )
        if self.model_size != other_reg.model_size:
            raise DimensionMismatchError(
                entered_name="the second regularization term",
                entered_dimension=other_reg.model_size,
                expected_source="the first regularization term",
                expected_dimension=self.model_size,
            )

        return CompositeRegularization(self, other=other_reg)

    def __rmul__(self, coefficient):
        r"""Multiply a regularization term with a constant number

        Parameters
        ----------
        coefficient : Number
            the first argument of "*" operator; must be a Number

        Returns
        -------
        BaseRegularization
            a regularization term ``resRegularization`` such that:

            - :math:`\text{resRegularization.reg}(m)=\text{coefficient}\times\text{self.reg}(m)`
            - :math:`\text{resRegularization.gradient}(m)=\text{coefficient}\times\text{self.gradient}(m)`
            - :math:`\text{resRegularization.hessian}(m)=\text{coefficient}\times\text{self.hessian}(m)`

        Raises
        ------
        TypeError
            when the ``coefficient`` is not of a python Number type

        Examples
        --------

        >>> from cofi import BaseProblem
        >>> from cofi.utils import QuadraticReg
        >>> reg = QuadraticReg(model_shape=(3,), weighting_matrix="damping")
        >>> my_problem = BaseProblem()
        >>> my_problem.set_regularization(10 * reg)

        """
        if not isinstance(coefficient, Number):
            raise TypeError(
                f"unsupported operand type(s) for *: '{coefficient.__class__.__name__}'"
                f" and '{self.__class__.__name__}"
            )
        return CompositeRegularization(self, coefficient=coefficient)


class CompositeRegularization(BaseRegularization):
    def __init__(self, base, other=None, coefficient=None):
        self.base = base
        self.other = other
        self.coefficient = coefficient

    @property
    def model_shape(self):
        return self.base.model_shape

    def reg(self, model):
        if self.other:
            return self.base.reg(model) + self.other.reg(model)
        elif self.coefficient is not None:
            return self.coefficient * self.base.reg(model)

    def gradient(self, model):
        if self.other:
            return self.base.gradient(model) + self.other.gradient(model)
        elif self.coefficient is not None:
            return self.coefficient * self.base.gradient(model)

    def hessian(self, model):
        if self.other:
            return self.base.hessian(model) + self.other.hessian(model)
        elif self.coefficient is not None:
            return self.coefficient * self.base.hessian(model)



import functools

from . import BaseInferenceTool, error_handler


# Official documentation for scipy.optimize.minimize
# https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html

# 'jac' will be used when choosing the following methods:
# CG, BFGS, Newton-CG, L-BFGS-B, TNC, SLSQP, dogleg, trust-ncg, trust-krylov, trust-exact
# and trust-constr

# 'hess' will be used when choosing the following methods:
# Newton-CG, dogleg, trust-ncg, trust-krylov, trust-exact and trust-constr

# 'hessp' will be used when choosing the following methods:
# Newton-CG, trust-ncg, trust-krylov, trust-constr

# 'bounds' will be used when choosing the following methods:
# Nelder-Mead, L-BFGS-B, TNC, SLSQP, Powell, and trust-constr

# 'constraints' will be used when choosing the following methods:
# COBYLA, SLSQP and trust-constr

# other arguments include: tol, options, callback


class ScipyOptMin(BaseInferenceTool):
    documentation_links = [
        "https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html"
    ]
    short_description = (
        "SciPy's optimizers that minimizes a scalar function with respect to "
        "one or more variables, check SciPy's documentation page for a list of methods"
    )

    @classmethod
    def required_in_problem(cls) -> set:
        return _init_class_methods()[0]

    @classmethod
    def optional_in_problem(cls) -> dict:
        return _init_class_methods()[1]

    @classmethod
    def required_in_options(cls) -> set:
        return _init_class_methods()[2]

    @classmethod
    def optional_in_options(cls) -> dict:
        return _init_class_methods()[3]

    @classmethod
    @functools.lru_cache(maxsize=None)
    def available_algorithms(cls) -> set:
        import scipy

        return set(scipy.optimize._minimize.MINIMIZE_METHODS)

    def __init__(self, inv_problem, inv_options):
        super().__init__(inv_problem, inv_options)
        self._components_used = list(self.required_in_problem())
        self._assign_args()

    def _assign_args(self):
        inv_problem = self.inv_problem
        # required_in_problem
        self._fun = inv_problem.objective
        self._x0 = inv_problem.initial_model
        # optional_in_problem
        _optional_in_problem_map = {
            "gradient": "jac",
            "hessian": "hess",
            "hessian_times_vector": "hessp",
            "bounds": "bounds",
            "constraints": "constraints",
        }
        defined_in_problem = self.inv_problem.defined_components()
        for component in _optional_in_problem_map:
            if component in defined_in_problem:
                self._params[_optional_in_problem_map[component]] = getattr(
                    self.inv_problem, component
                )
                self._components_used.append(component)
            else:  # default
                self._params[_optional_in_problem_map[component]] = (
                    self.optional_in_problem()[component]
                )

    def __call__(self) -> dict:
        opt_result = self._call_np_minimize()
        result = dict(opt_result.items())
        result["model"] = result.pop("x")
        return result

    @error_handler(
        when="when solving the optimization problem",
        context="calling `scipy.optimize.minimize`",
    )
    def _call_np_minimize(self):
        from scipy.optimize import minimize

        return minimize(
            fun=self._fun,
            x0=self._x0,
            args=(),  # handled by cofi.BaseProblem
            method=self._params["method"],
            jac=self._params["jac"],
            hess=self._params["hess"],
            hessp=self._params["hessp"],
            bounds=self._params["bounds"],
            constraints=self._params["constraints"],
            tol=self._params["tol"],
            callback=self._params["callback"],
            options=self._params["options"],
        )


@functools.lru_cache(maxsize=None)
def _init_class_methods():
    """get a list of arguments and defaults for scipy.minimize.minimize"""
    import inspect
    from scipy.optimize import minimize

    _scipy_minimize_args = dict(inspect.signature(minimize).parameters)
    _scipy_minimize_args["gradient"] = _scipy_minimize_args.pop("jac")
    _scipy_minimize_args["hessian"] = _scipy_minimize_args.pop("hess")
    _scipy_minimize_args["hessian_times_vector"] = _scipy_minimize_args.pop("hessp")
    components_used: list = []
    required_in_problem = {"objective", "initial_model"}  # `fun`, `x0`
    optional_in_problem = {
        k: v.default
        for k, v in _scipy_minimize_args.items()
        if k
        in {
            "gradient",
            "hessian",
            "hessian_times_vector",
            "bounds",
            "constraints",
        }
    }
    required_in_options = set()
    optional_in_options = {
        k: v.default
        for k, v in _scipy_minimize_args.items()
        if k in {"method", "tol", "callback", "options"}
    }
    return (
        required_in_problem,
        optional_in_problem,
        required_in_options,
        optional_in_options,
    )


# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> Nelder-Mead
# description: Uses the Simplex algorithm.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-neldermead.html#optimize-minimize-neldermead

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> Powell
# description: A modification of Powell’s method which is a conjugate direction method.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-powell.html#optimize-minimize-powell

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> CG
# description: Uses a nonlinear conjugate gradient algorithm by Polak and Ribiere, a variant of the Fletcher-Reeves method
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-cg.html#optimize-minimize-cg

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> BFGS
# description: Uses the quasi-Newton method of Broyden, Fletcher, Goldfarb, and Shanno (BFGS).
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-bfgs.html#optimize-minimize-bfgs

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> Newton-CG
# description: Uses a Newton-CG algorithm (also known as the truncated Newton method).
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-newtoncg.html#optimize-minimize-newtoncg

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> L-BFGS-B
# description: Uses the L-BFGS-B algorithm for bound constrained minimization.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-lbfgsb.html#optimize-minimize-lbfgsb

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> TNC
# description: Uses a truncated Newton algorithm to minimize a function with variables subject to bounds.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-tnc.html#optimize-minimize-tnc

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> COBYLA
# description: Uses the Constrained Optimization BY Linear Approximation (COBYLA) method.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-cobyla.html#optimize-minimize-cobyla

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> SLSQP
# description: Uses Sequential Least SQuares Programming to minimize a function of several variables with any combination of bounds, equality and inequality constraints.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-slsqp.html#optimize-minimize-slsqp

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> trust-constr
# description: A trust-region algorithm for constrained optimization.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-trustconstr.html#optimize-minimize-trustconstr

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> dogleg
# description: Uses the dog-leg trust-region algorithm for unconstrained minimization.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-dogleg.html#optimize-minimize-dogleg

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> trust-ncg
# description: Uses the Newton conjugate gradient trust-region algorithm for unconstrained minimization.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-trustncg.html#optimize-minimize-trustncg

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> trust-exact
# description: A trust-region method for unconstrained minimization in which quadratic subproblems are solved almost exactly.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-trustexact.html#optimize-minimize-trustexact

# CoFI -> Parameter estimation -> Optimization -> Non linear -> scipy.optimize.minimize -> trust-krylov
# description: Uses the Newton GLTR trust-region algorithm for unconstrained minimization.
# documentation: https://docs.scipy.org/doc/scipy/reference/optimize.minimize-trustkrylov.html#optimize-minimize-trustkrylov



import sys as _sys

from ._base_problem import BaseProblem
from ._inversion_options import InversionOptions
from ._inversion import Inversion, InversionResult, SamplingResult

from . import utils
from . import tools

from ._version import __version__


__all__ = [
    "BaseProblem",  # public API, basic usage
    "InversionOptions",  # public API, basic usage
    "Inversion",  # public API, basic usage
    "InversionResult",  # public API, for advanced usage (own solver)
    "SamplingResult",  # public API, for advanced usage (own solver)
]


# Set default logging handler to avoid "No handler found" warnings.
import logging as _logging

_logging.getLogger(__name__).addHandler(_logging.NullHandler())

# alias for deprecated API
_sys.modules["cofi.solvers"] = tools



from numbers import Number
from typing import Union, Tuple
import numpy as np

from ._reg_base import BaseRegularization
from .._exceptions import DimensionMismatchError


class ModelCovariance(BaseRegularization):
    r"""CoFI's utility abstract class to calculate model prior distribution

    See :class:`GaussianPrior` for a concrete subclass example.
    """


class GaussianPrior(ModelCovariance):
    r"""CoFI's utility class to calculate the Gaussian prior, given the inverse of
    model covariance matrix and the mean model

    :math:`GaussianPrior(C_m^{-1}, \mu) = (m-\mu)^TC_m^{-1}(m-\mu)`

    With gradient of:

    :math:`2\times C_m^{-1} (m-\mu)`

    And hessian of:

    :math:`2\times C_m^{-1}`

    Optionally, the inverse model covariance matrix can be constructed given the
    correlation lengths and sigma.

    Parameters
    ----------
    model_covariance_inv: np.ndarray or (tuple, float)
        the inverse model covariance matrix, either provided by users or specified so
        can be generated, corresponding to :math:`C_m^{-1}` in the formula above. When
        this is a tuple (i.e. CoFI will construct the matrix for you), it should be in
        the form of `(corr_lengths, sigma)`
    mean_model: np.ndarray
        the prior model, corresponding to :math:`\mu` in the formula above

    Raises
    ------
    TypeError
        if arguments aren't in accepted types as described above in "Parameters"
        section
    ValueError
        if shape of model_covariance_inv and mean_model doesn't match, or if shape of
        corr_lengths and mean_model doesn't match

    Examples
    --------

    Generate a Gaussian Prior term for models of size (4,4), with correlation lengths
    of (2,2), and sigma of 0.5:

    >>> from cofi.utils import GaussianPrior
    >>> import numpy as np
    >>> my_prior_model = np.array([[1,2],[3,4]])
    >>> my_reg = GaussianPrior(model_covariance_inv=((2,2), 0.5), mean_model=my_prior_model)
    >>> my_reg(np.array([[0,2],[1,0]]))
    101.53804950804782

    To use together with :class:`cofi.BaseProblem`:

    >>> from cofi import BaseProblem
    >>> my_problem = BaseProblem()
    >>> my_problem.set_regularization(my_reg)
    """

    def __init__(
        self,
        model_covariance_inv: Union[np.ndarray, Tuple[Tuple, float]],
        mean_model: np.ndarray,
    ):
        self._mu = np.ravel(mean_model)
        self._model_shape = mean_model.shape
        self._prepare_covariance_matrix_inv(model_covariance_inv, mean_model)

    def reg(self, model: np.ndarray) -> Number:
        flat_m = self._validate_model(model)
        diff_m = flat_m - self._mu
        return diff_m.T @ self._Cminv @ diff_m

    def gradient(self, model: np.ndarray) -> np.ndarray:
        flat_m = self._validate_model(model)
        return 2 * self._Cminv @ (flat_m - self._mu)

    def hessian(self, model: np.ndarray) -> np.ndarray:
        return 2 * self._Cminv

    @property
    def model_shape(self) -> tuple:
        return self._model_shape

    @property
    def gaussian_model_covariance_inv(self) -> np.ndarray:
        return self._Cminv

    def _prepare_covariance_matrix_inv(self, model_covariance_inv, mean_model):
        if isinstance(model_covariance_inv, np.ndarray):
            mu = self._mu
            Cminv = model_covariance_inv
            if Cminv.shape != (mu.shape[0], mu.shape[0]):
                raise ValueError(
                    f"({(mu.shape[0], mu.shape[0])}) expected for the shape of "
                    f"model_covariance_inv but got matrix of shape {Cminv.shape}"
                )
            self._Cminv = model_covariance_inv
        elif isinstance(model_covariance_inv, (tuple, list)):
            self._Cminv = self._generate_covariance_matrix_inv(
                mean_model.shape,
                model_covariance_inv[0],
                model_covariance_inv[1],
            )
        else:
            raise TypeError(
                "numpy.ndarray or (tuple, float) expected for `model_covariance_inv` "
                f"but got {model_covariance_inv} of type {type(model_covariance_inv)}"
            )

    def _generate_covariance_matrix_inv(
        self, model_shape: tuple, corr_lengths: tuple, sigma: float
    ):
        # ensure model_shape and corr_lengths have the same length
        if len(model_shape) != len(corr_lengths):
            raise ValueError(
                "`model_shape` and `corr_lengths` should have the same lengths, "
                f"but got {len(model_shape)} and {len(corr_lengths)}"
            )
        # generate grid of points for each dimension
        grids = np.meshgrid(*[np.arange(dim) for dim in model_shape], indexing="ij")
        # calculate distances between points for each pair of dimensions
        d_squared = sum([
            (grid.ravel()[None, :] - grid.ravel()[:, None]) ** 2 / corr_length**2
            for grid, corr_length in zip(grids, corr_lengths)
        ])
        # construct correlation matrix
        Cp = np.exp(-np.sqrt(d_squared))
        # construct variance matrix
        Sc = np.zeros((np.prod(model_shape), np.prod(model_shape)))
        np.fill_diagonal(Sc, sigma)
        # calculate covariance matrix
        covariance_matrix = Sc @ Cp @ Sc
        # calculate inverse covariance matrix
        covariance_matrix_inv = np.linalg.inv(covariance_matrix)
        return covariance_matrix_inv

    def _validate_model(self, model):
        flat_m = np.ravel(model)
        if flat_m.size != self.model_size:
            raise DimensionMismatchError(
                entered_name="model",
                entered_dimension=model.shape,
                expected_source="model_size",
                expected_dimension=self.model_size,
            )
        return flat_m



from numbers import Number
from typing import Callable, Union, Tuple, Sequence
import functools
import json

import numpy as np
import scipy

from ._exceptions import (
    DimensionMismatchError,
    InvalidOptionError,
    InvocationError,
    NotDefinedError,
)


class BaseProblem:
    r"""Base class for an inversion problem setup.

    An inversion problem can be defined in different ways, depending on the level of
    flexibility or control in ways you'd like to evaluate a model, as well as the
    solving approaches you'd like to apply on the problem.

    To define an inversion problem that is intended to be solved by **parameter estimation**,
    you may consider setting the following functions or properties:

    - ``objective`` function, or
    - ``data_misfit`` function plus ``regularization`` function
    - ``data_misfit="L2"``, ``data``, ``forward`` and ``regularization`` function
    - In addition, it can sometimes be helpful (e.g. increase the speed of inversion)
      to define more things in a ``BaseProblem`` object: ``gradient`` of objective
      function, ``residual`` vector, ``jacobian`` of forward function, etc.

    To define an inversion problem that is intended to be solved by **ensemble methods**
    (work in progress),
    you may consider setting the following functions or properties:

    - ``log_posterier`` function, or
    - ``log_likelihood`` and ``log_prior`` functions

    .. TBD: we will also add support for ``bounds`` and ``constraints`` as a part of
    .. ``BaseProblem`` definition.

    Here is a complete list of how we would deduce from existing information about the
    ``BaseProblem`` object you've defined:

    .. list-table:: Table: user defined -> we generate for you
        :widths: 35 35 30
        :header-rows: 1

        * - what you define
          - what we generate for you
          - examples
        * - ``data_misfit``
          - ``objective`` (assuming there's no regularization)
          - (work in progress)
        * - ``data_misfit``, ``regularization``
          - ``objective``
          - `linear regression (optimizer) <https://github.com/inlab-geo/cofi-examples/blob/main/notebooks/linear_regression/linear_regression_optimizer_minimize.py>`_
        * - ``forward``, ``data``
          - ``residual``
          - (work in progress)
        * - ``hessian``
          - ``hessian_times_vector``
          - (work in progress)
        * - ``jacobian``
          - ``jacobian_times_vector``
          - `linear regression (linear system solver) <https://github.com/inlab-geo/cofi-examples/blob/main/notebooks/linear_regression/linear_regression_linear_system_solver.py>`_
        * - ``log_prior``, ``log_likelihood``
          - ``log_posterior``
          - `linear regression (sampler) <https://github.com/inlab-geo/cofi-examples/blob/main/notebooks/linear_regression/linear_regression_emcee_sampler.py>`_


    .. admonition:: One quick example of BaseProblem
       :class: dropdown, attention

        >>> from cofi import BaseProblem
        >>> import numpy as np
        >>> inv_problem = BaseProblem()
        >>> data_x = np.array([1, 2, 3, 4])
        >>> data_y = np.array([3.2, 3.9, 5.1, 6.2])
        >>> def my_forward(model):
        ...   assert len(model) == 2
        ...   return model[0] + model[1] * data_x
        ...
        >>> inv_problem.set_data(data_y)
        >>> inv_problem.set_forward(my_forward)
        >>> inv_problem.set_data_misfit("L2")
        >>> inv_problem.summary()
        Summary for inversion problem: BaseProblem
        =====================================================================
        Model shape: Unknown
        ---------------------------------------------------------------------
        List of functions/properties set by you:
        ['forward', 'data']
        ---------------------------------------------------------------------
        List of functions/properties created based on what you have provided:
        ['objective', 'residual', 'data_misfit']
        ( Note that you did not set regularization )
        ---------------------------------------------------------------------
        List of functions/properties not set by you:
        ['objective', 'gradient', 'hessian', 'hessian_times_vector', 'residual',
        'jacobian', 'jacobian_times_vector', 'data_misfit', 'regularization',
        'initial_model', 'model_shape', 'bounds', 'constraints']

    .. tip::

        Check :ref:`Set Methods <set_methods>` to see a full list of methods to attach
        information about the problem.

    Some blocks above may be deduced from other existing information. For instance,
    once you've defined your data, forward operator and how you'd like to calculate
    the data misfit, we are able to generate ``data_misfit`` and ``residual`` for you.
    The ``summary()`` method prints what blocks you've defined, what are not yet defined,
    and what are generated automatically for you.

    At any point of defining your inversion problem, the ``suggest_tools()``
    method helps get a list of inference tools that can be applied to your problem based on
    what have been supplied so far.

    .. tip::

        :ref:`Helper Methods <helper_methods>` are there to help you illustrate what's in your
        ``BaseProblem`` object.

        Additionally, :ref:`Properties/Functaions <prop_func>` set by you are accessible
        through the ``BaseProblem`` object directly.

    :ref:`back to top <top_BaseProblem>`

    .. _set_methods:

    .. rubric:: Set Methods

    Here are a series of ``set`` methods:

    .. autosummary::
        BaseProblem.set_objective
        BaseProblem.set_log_posterior
        BaseProblem.set_log_posterior_with_blobs
        BaseProblem.set_blobs_dtype
        BaseProblem.set_log_likelihood
        BaseProblem.set_log_prior
        BaseProblem.set_gradient
        BaseProblem.set_hessian
        BaseProblem.set_hessian_times_vector
        BaseProblem.set_residual
        BaseProblem.set_jacobian
        BaseProblem.set_jacobian_times_vector
        BaseProblem.set_data_misfit
        BaseProblem.set_regularization
        BaseProblem.set_forward
        BaseProblem.set_data
        BaseProblem.set_data_covariance
        BaseProblem.set_data_covariance_inv
        BaseProblem.set_data_from_file
        BaseProblem.set_initial_model
        BaseProblem.set_model_shape
        .. BaseProblem.set_bounds
        .. BaseProblem.set_constraints

    :ref:`back to top <top_BaseProblem>`

    .. _helper_methods:

    .. rubric:: Helper Methods

    Here are helper methods that check what you've defined to the ``BaseProblem``:

    .. autosummary::

        BaseProblem.summary
        BaseProblem.suggest_tools
        BaseProblem.defined_components

    :ref:`back to top <top_BaseProblem>`

    .. _prop_func:

    .. rubric:: Properties/Functions of the Problem

    In case you'd like to check, the properties/functions defined using the ``set``
    methods above are attached directly to ``BaseProblem`` and can be accessed:

    .. autosummary::

        BaseProblem.objective
        BaseProblem.log_posterior
        BaseProblem.log_posterior_with_blobs
        BaseProblem.log_likelihood
        BaseProblem.log_prior
        BaseProblem.gradient
        BaseProblem.hessian
        BaseProblem.hessian_times_vector
        BaseProblem.residual
        BaseProblem.jacobian
        BaseProblem.jacobian_times_vector
        BaseProblem.data_misfit
        BaseProblem.regularization
        BaseProblem.regularization_matrix
        BaseProblem.forward
        BaseProblem.name
        BaseProblem.data
        BaseProblem.data_covariance
        BaseProblem.data_covariance_inv
        BaseProblem.model_covariance
        BaseProblem.model_covariance_inv
        BaseProblem.initial_model
        BaseProblem.model_shape
        BaseProblem.blobs_dtype
        BaseProblem.bounds
        BaseProblem.constraints

    :ref:`back to top <top_BaseProblem>`

    """

    all_components = [
        "objective",
        "log_posterior",
        "log_posterior_with_blobs",
        "log_likelihood",
        "log_prior",
        "gradient",
        "hessian",
        "hessian_times_vector",
        "residual",
        "jacobian",
        "jacobian_times_vector",
        "data_misfit",
        "regularization",
        "regularization_matrix",
        "forward",
        "data",
        "data_covariance",
        "data_covariance_inv",
        "initial_model",
        "model_shape",
        "blobs_dtype",
        "bounds",
        "constraints",
    ]

    def __init__(self, **kwargs):
        for kw, val in kwargs.items():
            if kw in self.all_components:
                set_func = getattr(self, f"set_{kw}")
                if isinstance(val, dict):
                    set_func(**val)
                else:
                    set_func(val)

    def objective(self, model: np.ndarray, *args, **kwargs) -> Number:
        """Method for computing the objective function given a model

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        Number
            The objective function value for the given model

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="objective")

    def log_posterior(self, model: np.ndarray, *args, **kwargs) -> Number:
        """Method for computing the log of posterior probability density given a model

        This is typically the sum of log prior and log likelihood.

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        Number
            the posterior probability density value

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="log_posterior")

    def log_posterior_with_blobs(
        self, model: np.ndarray, *args, **kwargs
    ) -> Tuple[Number]:
        """Method for computing the log of posterior probability density and related
        information given a model

        The "related information" can be defined by you
        (via :meth:`set_log_posterior_with_blobs`), but they will only be
        stored properly when you perform sampling with ``emcee``.

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        Tuple[Number]
            the posterior probability density value, and other information you've set to
            return together with the former

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="log_posterior_with_blobs")

    def log_prior(self, model: np.ndarray, *args, **kwargs) -> Number:
        """Method for computing the log of prior probability density given a model

        This reflects your prior belief about the model distribution.

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        Number
            the prior probability density value

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="log_prior")

    def log_likelihood(self, model: np.ndarray, *args, **kwargs) -> Number:
        """Method for computing the log of likelihood probability density given a model

        This reflects the probability distribution of the observations given the model.

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        Number
            the likelihood probability density value

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="log_likelihood")

    def gradient(self, model: np.ndarray, *args, **kwargs) -> np.ndarray:
        """Method for computing the gradient of objective function with respect to model, given a model

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        np.ndarray
            the gradient (first derivative) of objective function with repect to the model

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="gradient")

    def hessian(self, model: np.ndarray, *args, **kwargs) -> np.ndarray:
        """Method for computing the Hessian of objective function with respect to model, given a model

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        np.ndarray
            the Hessian (second derivative) of objective function with respect to the model

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="hessian")

    def hessian_times_vector(
        self, model: np.ndarray, vector: np.ndarray, *args, **kwargs
    ) -> np.ndarray:
        """Method for computing the dot product of the Hessian and an arbitrary vector, given a model

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate
        vector : np.ndarray
            an arbitrary vector

        Returns
        -------
        np.ndarray
            Hessian times an arbitrary vector

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="hessian_times_vector")

    def residual(self, model: np.ndarray, *args, **kwargs) -> np.ndarray:
        r"""Method for computing the residual vector given a model.

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        np.ndarray
            the residual vector, :math:`\text{forward}(\text{model})-\text{observations}`

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="residual")

    def jacobian(self, model: np.ndarray, *args, **kwargs) -> np.ndarray:
        r"""method for computing the jacobian of forward function with respect to model, given a model

        parameters
        ----------
        model : np.ndarray
            a model to evaluate

        returns
        -------
        np.ndarray
            the jacobian matrix, :math:`\frac{\partial{\text{forward}(\text{model})}}{\partial\text{model}}`

        raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="jacobian")

    def jacobian_times_vector(
        self, model: np.ndarray, vector: np.ndarray, *args, **kwargs
    ) -> np.ndarray:
        """Method for computing the dot product of the Jacobian and an arbitrary vector, given a model

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate
        vector : np.ndarray
            an arbitrary vector

        Returns
        -------
        np.ndarray
            the Jacobian matrix times the given vector

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="jacobian_times_vector")

    def data_misfit(self, model: np.ndarray, *args, **kwargs) -> Number:
        """Method for computing the data misfit value given a model

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        Number
            the data misfit evaluated based on how you've defined it

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="data_misfit")

    def regularization(self, model: np.ndarray, *args, **kwargs) -> Number:
        """Method for computing the regularization value given a model

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        Number
            the regularization value evaluated based on how you've defined it

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="regularization")

    def regularization_matrix(self, model: np.ndarray, *args, **kwargs) -> np.ndarray:
        """Method for computing the regularization weighting matrix

        Parameters
        ----------
        model : np.ndarray
            a model that helps calculate regularization matrix. In most cases this is
            not needed, but you have the flexibility to set this as a function

        Returns
        -------
        np.ndarray
            the regularization matrix of dimension ``(model_size, model_size)``

        Raises
        ------
        NotDefinedError
            when this method is not set
        """
        raise NotDefinedError(needs="regularization_matrix")

    def forward(self, model: np.ndarray, *args, **kwargs) -> Union[np.ndarray, Number]:
        """Method to perform the forward operation given a model

        Parameters
        ----------
        model : np.ndarray
            a model to evaluate

        Returns
        -------
        np.ndarray or Number
            the synthetics data

        Raises
        ------
        NotDefinedError
            when this method is not set and cannot be generated from known information
        """
        raise NotDefinedError(needs="forward")

    # TO ADD a set method, remember to do the following:
    # - def set_something(self, something)
    # - def something(self), this is a property / function
    # - def something_defined(self) -> bool
    # - add checking to self.dall_components
    # - add `set_something` and `something` to documentation list on top of this file
    # - check if there's anything to add to autogen_table
    # - add tests in tests/test_base_problem.py ("test_non_set", etc.)

    def set_objective(
        self,
        obj_func: Callable[[np.ndarray], Number],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the objective function to minimize

        Alternatively, objective function can be set implicitly (computed by us) if one of
        the following combinations is set:

        - :meth:`set_data_misfit` + :meth:`set_regularization`
        - :meth:`set_data_misfit` (in this case, regularization is default
          to 0)

        Parameters
        ----------
        obj_func : Callable[[np.ndarray], Number]
            the objective function that matches :meth:`objective` in
            signature
        args : list, optional
            extra list of positional arguments for the objective function
        kwargs : dict, optional
            extra dict of keyword arguments for the objective function
        """
        self.objective = _FunctionWrapper("objective", obj_func, args, kwargs)
        self._update_autogen("objective")

    def set_log_posterior(
        self,
        log_posterior_func: Callable[[np.ndarray], Number],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the log of posterior probability density

        Alternatively, log_posterior function can be set implicitly (computed by us) if
        :func:`set_log_prior` and :func:`set_log_likelihood` are defined.

        Parameters
        ----------
        log_posterior_func : Callable[[np.ndarray], Number]
            the log_posterior function that matches :meth:`log_posterior`
            in signature
        args : list, optional
            extra list of positional arguments for log_posterior function
        kwargs : dict, optional
            extra dict of keyword arguments for log_posterior function

        """
        self.log_posterior = _FunctionWrapper(
            "log_posterior", log_posterior_func, args, kwargs
        )
        self._update_autogen("log_posterior")

    def set_log_posterior_with_blobs(
        self,
        log_posterior_blobs_func: Callable[[np.ndarray], Tuple[Number]],
        blobs_dtype=None,
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function that computes the log of posterior prabability density
        and returns extra information along with log posterior

        The extra blobs returned will only get used when you are using ``emcee`` to
        sample the posterior distribution. Check
        `this emcee documentation page <https://emcee.readthedocs.io/en/stable/user/blobs/>`_
        to understand what blobs are.

        If you use other backend samplers, you can still set ``log_posterior`` using
        this function, and we will generate :meth:`log_posterior` to return
        only the first output from :meth:`log_posterior_with_blobs`.

        This method is also generated automatically by us if you've defined both
        :meth:`log_prior` and :meth:`log_likelihood`. In that
        case, the ``blobs_dtype`` is set to be
        ``[("log_likelihood", float), ("log_prior", float)]``.

        Parameters
        ----------
        log_posterior_blobs_func : Callable[[np.ndarray], Tuple[Number]
            the log_posterior_with_blobs function that matches
            :meth:`log_posterior_blobs_func` in signature
        blobs_dtype : list, optional
            a list of tuples that specify the names and type of the blobs, e.g.
            ``[("log_likelihood", float), ("log_prior", float)]``. If not set, the
            blobs will still be recorded during sampling in the order they are
            returned from :meth:`log_posterior_blobs_func`
        args : list, optional
            extra list of positional arguments for log_posterior function
        kwargs : dict, optional
            extra dict of keyword arguments for log_posterior function
        """
        self.log_posterior_with_blobs = _FunctionWrapper(
            "log_posterior_with_blobs", log_posterior_blobs_func, args, kwargs
        )
        self._update_autogen("log_posterior_with_blobs")
        if blobs_dtype:
            self._blobs_dtype = blobs_dtype

    def set_blobs_dtype(self, blobs_dtype: list):
        r"""Sets the name and type for the extra information you'd like to calculate on
        each sampling step

        This only gets used when you are using ``emcee`` to sample the posterior
        distribution. Check `this emcee documentation page <https://emcee.readthedocs.io/en/stable/user/blobs/>`_
        to understand what blobs are.

        Parameters
        ----------
        blobs_dtype : list
            a list of tuples that specify the names and type of the blobs, e.g.
            ``[("log_likelihood", float), ("log_prior", float)]``
        """
        self._blobs_dtype = blobs_dtype
        self._update_autogen("blobs_dtype")

    def set_log_prior(
        self,
        log_prior_func: Callable[[np.ndarray], Number],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the log of prior probability density

        Parameters
        ----------
        log_prior_func : Callable[[np.ndarray], Number]
            the log_prior function that matches :meth:`log_prior`
            in signature
        args : list, optional
            extra list of positional arguments for log_prior function
        kwargs : dict, optional
            extra dict of keyword arguments for log_prior function
        """
        self.log_prior = _FunctionWrapper("log_prior", log_prior_func, args, kwargs)
        self._update_autogen("log_prior")

    def set_log_likelihood(
        self,
        log_likelihood_func: Callable[[np.ndarray], Number],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the log of likelihood probability density

        Parameters
        ----------
        log_likelihood_func : Callable[[np.ndarray], Number]
            the log_likelihood function that matches :meth:`log_likelihood`
            in signature
        args : list, optional
            extra list of positional arguments for log_likelihood function
        kwargs : dict, optional
            extra dict of keyword arguments for log_likelihood function
        """
        self.log_likelihood = _FunctionWrapper(
            "log_likelihood", log_likelihood_func, args, kwargs
        )
        self._update_autogen("log_likelihood")

    def set_gradient(
        self,
        grad_func: Callable[[np.ndarray], np.ndarray],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the gradient of objective function w.r.t the
        model

        Parameters
        ----------
        obj_func : Callable[[np.ndarray], Number]
            the gradient function that matches :meth:`gradient` in
            signature
        args : list, optional
            extra list of positional arguments for gradient function
        kwargs : dict, optional
            extra dict of keyword arguments for gradient function
        """
        self.gradient = _FunctionWrapper("gradient", grad_func, args, kwargs)
        self._update_autogen("gradient")

    def set_hessian(
        self,
        hess_func: Union[Callable[[np.ndarray], np.ndarray], np.ndarray],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the Hessian of objective function w.r.t the
        model

        Parameters
        ----------
        hess_func : (function - np.ndarray -> np.ndarray) or np.ndarray
            the Hessian function that matches :meth:`hessian` in
            signature. Alternatively, provide a matrix if the Hessian is a constant.
        args : list, optional
            extra list of positional arguments for hessian function
        kwargs : dict, optional
            extra dict of keyword arguments for hessian function
        """
        if isinstance(hess_func, np.ndarray):
            self.hessian = _FunctionWrapper(
                "hessian", _matrix_to_func, args=[hess_func]
            )
        else:
            self.hessian = _FunctionWrapper("hessian", hess_func, args, kwargs)
        self._update_autogen("hessian")

    def set_hessian_times_vector(
        self,
        hess_vec_func: Callable[[np.ndarray, np.ndarray], np.ndarray],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the Hessian (of objective function) times
        an arbitrary vector

        Alternatively, hessian_times_vector function can be set implicitly (computed by us)
        if :func:`set_hessian` is defined.

        Parameters
        ----------
        hess_vec_func : Callable[[np.ndarray, np.ndarray], np.ndarray]
            the function that computes the product of Hessian and an arbitrary vector,
            in the same signature as :meth:`hessian_times_vector`
        args : list, optional
            extra list of positional arguments for hessian_times_vector function
        kwargs : dict, optional
            extra dict of keyword arguments for hessian_times_vector function
        """
        self.hessian_times_vector = _FunctionWrapper(
            "hessian_times_vector", hess_vec_func, args, kwargs
        )
        self._update_autogen("hessian_times_vector")

    def set_residual(
        self,
        res_func: Callable[[np.ndarray], np.ndarray],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the residual vector/matrix

        Alternatively, residual function can be set implicitly (computed by us)
        if both :func:`set_forward` and data (:func:`set_data` or
        :func:`set_data_from_file`) are defined.

        Parameters
        ----------
        res_func : Callable[[np.ndarray], np.ndarray]
            the residual function that matches :meth:`residual` in
            signature
        args : list, optional
            extra list of positional arguments for residual function
        kwargs : dict, optional
            extra dict of keyword arguments for residual function
        """
        self.residual = _FunctionWrapper("residual", res_func, args, kwargs)
        self._update_autogen("residual")

    def set_jacobian(
        self,
        jac_func: Union[Callable[[np.ndarray], np.ndarray], np.ndarray],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the Jacobian matrix, namely first
        derivative of forward function with respect to the model

        Parameters
        ----------
        jac_func : (function - np.ndarray -> np.ndarray) or np.ndarray
            the Jacobian function that matches :meth:`residual` in
            signature. Alternatively, provide a matrix if the Jacobian is a constant.
        args : list, optional
            extra list of positional arguments for jacobian function
        kwargs : dict, optional
            extra dict of keyword arguments for jacobian function
        """
        if isinstance(jac_func, np.ndarray) or scipy.sparse.issparse(jac_func):
            self.jacobian = _FunctionWrapper(
                "jacobian", _matrix_to_func, args=[jac_func]
            )
        else:
            self.jacobian = _FunctionWrapper("jacobian", jac_func, args, kwargs)
        self._update_autogen("jacobian")

    def set_jacobian_times_vector(
        self,
        jac_vec_func: Callable[[np.ndarray, np.ndarray], np.ndarray],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the Jacobian (of forward function) times
        an arbitrary vector

        Alternatively, jacobian_times_vector function can be set implicitly (computed by us)
        if :func:`set_jacobian` is defined.

        Parameters
        ----------
        jac_vec_func : Callable[[np.ndarray, np.ndarray], np.ndarray]
            the function that computes the product of Jacobian and an arbitrary vector,
            in the same signature as :meth:`jacobian_times_vector`
        args : list, optional
            extra list of positional arguments for jacobian_times_vector function
        kwargs : dict, optional
            extra dict of keyword arguments for jacobian_times_vector function
        """
        self.jacobian_times_vector = _FunctionWrapper(
            "jacobian_times_vector", jac_vec_func, args, kwargs
        )
        self._update_autogen("jacobian_times_vector")

    def set_data_misfit(
        self,
        data_misfit: Union[str, Callable[[np.ndarray], Number]],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the data misfit

        You can either pass in a custom function or a short string that describes the
        data misfit function (e.g. ``"squared error"``)

        If you choose ``data_misfit="squared error"``, and:

        - If you have :meth:`residual` defined, or :meth:`data` and :meth:`forward`
          defined, then :math:`\text{data_misfit}=\text{residual}^T \text{residual}`

          - where :math:`\text{residual}=\text{forward}(\text{model})-\text{observations}`

        - If you **additionally** have :meth:`data_covariance_inv` defined, then
          :math:`\text{data_misfit}=\text{residual}^TC_d^{-1}\text{residual}`

          - where :math:`C_d^{-1}=\text{data_covariance_inv}`

        - Otherwise you might face an error when actually calling the
          :meth:`data_misfit` method.

        Alternatively, pass in your own data misfit function (or objective function
        directly through :meth:`set_objective`).

        Parameters
        ----------
        data_misfit : str or (function - np.ndarray -> Number)
            either ``"squared error"``, or a data misfit function that matches
            :meth:`data_misfit` in signature.
        args : list, optional
            extra list of positional arguments for data_misfit function
        kwargs : dict, optional
            extra dict of keyword arguments for data_misfit function

        Raises
        ------
        InvalidOptionError
            when you've passed in a string not in our supported data misfit list
        """
        if isinstance(data_misfit, str):
            # if we have more options later, handle in same way as set_regularization
            if data_misfit in [
                "least squares",
                "least square",
                "squared error",
            ]:
                self.data_misfit = _FunctionWrapper(
                    "data_misfit", self._data_misfit_squared_error, autogen=True
                )
            else:
                raise InvalidOptionError(
                    name="data misfit",
                    invalid_option=data_misfit,
                    valid_options=["least squares"],
                )
        else:
            self.data_misfit = _FunctionWrapper(
                "data_misfit", data_misfit, args, kwargs
            )
        self._update_autogen("data_misfit")

    def set_regularization(
        self,
        regularization: Callable[[np.ndarray], Number] = None,
        regularization_matrix: Union[
            np.ndarray, Callable[[np.ndarray], np.ndarray]
        ] = None,
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to compute the regularization

        You can either pass in a custom function or a string/number that describes the
        order of the norm. We use :func:`numpy.linalg.norm` as our backend
        implementation, so the order can be chosen from:

        Parameters
        ----------
        regularization : function - np.ndarray -> Number
            a regularization function that matches :meth:`regularization` in signature.
        regularization_matrix : np.ndarray or (function - np.ndarray -> np.ndarray)
            a matrix of shape ``(model_size, model_size)``, or a function that takes in
            a model and calculates the (weighting) matrix.

            - If this is None,
              :math:`\text{regularization}(model)=\lambda\times\text{regularization}(model)`
            - If this is set to be a matrix (np.ndarray, or other array like types),
              :math:`\text{regularization}(model)=\lambda\times\text{regularization}(\text{regularization_matrix}\cdot model)`
            - If this is set to be a function that returns a matrix,
              :math:`\text{regularization}(model)=\lambda\times\text{regularization}(\text{regularization_matrix}(model)\cdot model)`

        args : list, optional
            extra list of positional arguments for regularization function
        kwargs : dict, optional
            extra dict of keyword arguments for regularization function

        Examples
        --------

        We demonstrate with a few examples on a ``BaseProblem`` instance.

        >>> from cofi import BaseProblem
        >>> inv_problem = BaseProblem()

        1. Example with a custom regularization function

        >>> inv_problem.set_regularization(lambda x: sum(x))
        >>> inv_problem.regularization([1,1])
        2

        2. Example with a custom regularization + a regularization matrix

        >>> inv_problem.set_regularization(lambda x: np.sum(x**2), np.eye(3))
        >>> inv_problem.regularization([1,1])
        2
        """
        if regularization is None and regularization_matrix is None:
            raise ValueError(
                "You must set either `regularization` or `regularization_matrix`"
            )
        # preprocess regularization_matrix if there is one
        _reg_matrix = None
        if regularization_matrix is not None:
            _reg_matrix = regularization_matrix
        elif regularization is not None and hasattr(regularization, "matrix"):
            _reg_matrix = regularization.matrix
        # wrap regularization_matrix as a function
        if _reg_matrix is not None and np.ndim(_reg_matrix) != 0:
            self.regularization_matrix = _FunctionWrapper(
                "regularization_matrix", _matrix_to_func, args=[_reg_matrix]
            )
        elif _reg_matrix is not None and callable(_reg_matrix):
            self.regularization_matrix = _FunctionWrapper(
                "regularization_matrix", _reg_matrix
            )
        else:
            self.regularization_matrix = None
        # process regularization function
        if regularization is not None:
            self.regularization = _FunctionWrapper(
                "regularization", regularization, args, kwargs
            )
        # update some autogenerated functions (as usual)
        self._update_autogen("regularization")

    def set_forward(
        self,
        forward: Callable[[np.ndarray], Union[np.ndarray, Number]],
        args: list = None,
        kwargs: dict = None,
    ):
        r"""Sets the function to perform the forward operation

        Parameters
        ----------
        forward : function - np.ndarray -> (np.ndarray or Number)
            the forward function that matches :meth:`forward` in signature
        args : list, optional
            extra list of positional arguments for forward function
        kwargs : dict, optional
            extra dict of keyword arguments for forward function
        """
        self.forward = _FunctionWrapper("forward", forward, args, kwargs)
        self._update_autogen("forward")

    def set_data(
        self,
        data_obs: np.ndarray,
        data_cov: np.ndarray = None,
        data_cov_inv: np.ndarray = None,
    ):
        """Sets the data observations and optionally data covariance matrix

        Parameters
        ----------
        data_obs : np.ndarray
            the observations
        data_cov : np.ndarray, optional
            the data covariance matrix that helps estimate uncertainty, with dimension
            (N,N) where N is the number of data points
        """
        self._data = data_obs
        self._update_autogen("data")
        if data_cov is not None:
            self.set_data_covariance(data_cov)
        if data_cov_inv is not None:
            self.set_data_covariance_inv(data_cov_inv)

    def set_data_covariance(self, data_cov: np.ndarray):
        """Sets the data covariance matrix to help estimate uncertainty

        Parameters
        ----------
        data_cov : np.ndarray
            the data covariance matrix, with dimension (N,N) where N is the number
            of data points
        """
        self._data_covariance = data_cov
        self._update_autogen("data_covariance")

    def set_data_covariance_inv(self, data_cov_inv: np.ndarray):
        """Sets the data covariance matrix to help estimate uncertainty

        Parameters
        ----------
        data_cov : np.ndarray
            the data covariance matrix, with dimension (N,N) where N is the number
            of data points
        """
        self._data_covariance_inv = data_cov_inv
        self._update_autogen("data_covariance_inv")

    def set_data_from_file(self, file_path, obs_idx=-1, data_cov: np.ndarray = None):
        r"""Sets the data for this problem from a give file path

        This function uses :func:`numpy.loadtxt` or :func:`numpy.load` to read
        data file, depending on the file type.

        Parameters
        ----------
        file_path : str
            a relative/absolute file path for the data
        obs_idx : int or list, optional
            the index/indices of observations within the data file, by default -1
        data_cov : np.ndarray, optional
            the data covariance matrix that helps estimate uncertainty, with dimension
            (N,N) where N is the number of data points
        """
        delimiter = None  # try to detect what delimiter is used
        if file_path.endswith(("npy", "npz")):
            data = np.load(file_path)
        elif file_path.endswith(("pickle", "pkl")):
            data = np.load(file_path, allow_pickle=True)
        else:
            with open(file_path) as file:
                first_line = file.readline()
                if "," in first_line:
                    delimiter = ","
            data = np.loadtxt(file_path, delimiter=delimiter)
        self.set_data(data[:, obs_idx], data_cov)

    def set_initial_model(self, init_model: np.ndarray):
        r"""Sets the starting point for the model

        Once set, we will infer the property :meth:`model_shape` in
        case this is required for some inference tools

        Parameters
        ----------
        init_model : np.ndarray
            the initial model
        """
        self._initial_model = init_model
        self._model_shape = init_model.shape if hasattr(init_model, "shape") else (1,)

    def set_model_shape(self, model_shape: Tuple):
        r"""Sets the model shape explicitly

        Parameters
        ----------
        model_shape : Tuple
            a tuple that describes model shape

        Raises
        ------
        DimensionMismatchError
            when you've defined an initial_model through :meth:`set_initial_model`
            but their shapes don't match
        """
        if self.initial_model_defined and self._model_shape != model_shape:
            try:
                np.reshape(self.initial_model, model_shape)
            except ValueError as err:
                raise DimensionMismatchError(
                    entered_dimension=model_shape,
                    entered_name="model shape",
                    expected_dimension=self.initial_model.shape,
                    expected_source="initial model",
                ) from err
        self._model_shape = model_shape

    def set_bounds(self, bounds: Sequence[Tuple[Number, Number]]):
        """TODO document me

        Parameters
        ----------
        bounds : Sequence[Tuple[Number, Number]]
            _description_
        """
        self._bounds = bounds

    def set_constraints(self, constraints):
        """TODO document me

        Parameters
        ----------
        constraints : _type_
            _description_
        """
        # TODO - what's the type of this? (ref: scipy has Constraint class)
        self._constraints = constraints

    def _defined_components(self, defined_only=True) -> Tuple[set, set]:
        _to_check = self.all_components
        defined = [
            func_name
            for func_name in _to_check
            if getattr(self, f"{func_name}_defined")
        ]
        if defined_only:
            return defined

        def _check_autogen(elem):
            _elem = getattr(self, elem)
            return isinstance(_elem, _FunctionWrapper) and _elem.autogen

        created_by_us = [elem for elem in defined if _check_autogen(elem)]
        return [elem for elem in defined if elem not in created_by_us], created_by_us

    def defined_components(self) -> set:
        r"""Returns a set of components that are defined for the ``BaseProblem`` object

        These include both the ones you've set explicitly through the :ref:`Set Methods <set_methods>`
        and the ones that are deduced from existing information.

        Returns
        -------
        set
            a set of strings describing what are defined
        """
        return self._defined_components()

    def suggest_tools(self, print_to_console=True) -> dict:
        r"""Prints / Returns the backend inversion tools that you can use, based on things
        defined for this ``BaseProblem`` instance, grouped by solving method

        Parameters
        ----------
        print_to_console : bool, optional
            if set to ``True``, this method will both print and return the dictionary
            of backend tools in a tree structure; if set to ``False``, then it will not
            print to console and will only return the dictionary; by default ``True``

        Returns
        -------
        dict
            a tree structure of solving methods we provide, with the leaf nodes being a
            list of backend inversion tools suggested based on what information you've
            provided to this ``BaseProblem`` object

        Examples
        --------

        .. admonition:: example usage for BaseProblem.suggest_tools()
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 6

                >>> from cofi import BaseProblem
                >>> import numpy as np
                >>> inv_problem = BaseProblem()
                >>> inv_problem.set_initial_model(np.array([1,2,3]))
                >>> inv_problem.set_data_misfit("least squares")
                >>> inv_problem.suggest_tools()
                Based on what you've provided so far, here are possible tools:
                {
                    "optimization": [
                        "scipy.optimize.minimize"
                    ],
                    "matrix solvers": [],
                    "sampling": []
                }

        """
        to_suggest = dict()
        all_components = self.defined_components()
        from .tools import inference_tools_table

        for solving_method in inference_tools_table:
            backend_tools = inference_tools_table[solving_method]
            to_suggest[solving_method] = []
            for tool in backend_tools:
                solver_class = backend_tools[tool]
                required = solver_class.required_in_problem()
                if required.issubset(all_components):
                    to_suggest[solving_method].append(tool)
        if print_to_console:
            print("Based on what you've provided so far, here are possible tools:")
            print(json.dumps(to_suggest, indent=4))
        return to_suggest

    @property
    def data(self) -> np.ndarray:
        r"""the observations, set by :meth:`set_data` or
        :meth:`set_data_from_file`

        Raises
        ------
        NotDefinedError
            when this property has not been defined by methods above
        """
        if hasattr(self, "_data") and self._data is not None:
            return self._data
        raise NotDefinedError(needs="data")

    @property
    def data_covariance(self) -> np.ndarray:
        """the data covariance matrix, set by :meth:`set_data_covariance`,
        :meth:`set_data` or :meth:`set_data_from_file`.

        Raises
        ------
        NotDefinedError
            when this property has not been defined by methods above
        """
        if hasattr(self, "_data_covariance") and self._data_covariance is not None:
            return self._data_covariance
        raise NotDefinedError(needs="data covariance matrix")

    @property
    def data_covariance_inv(self) -> np.ndarray:
        """the data covariance matrix, set by :meth:`set_data_covariance_inv`,
        :meth:`set_data` or :meth:`set_data_from_file`.

        Raises
        ------
        NotDefinedError
            when this property has not been defined by methods above
        """
        if (
            hasattr(self, "_data_covariance_inv")
            and self._data_covariance_inv is not None
        ):
            return self._data_covariance_inv
        raise NotDefinedError(needs="inverse data covariance matrix")

    @property
    def initial_model(self) -> np.ndarray:
        r"""the initial model, needed for some iterative optimization tools that
        requires a starting point

        Raises
        ------
        NotDefinedError
            when this property has not been defined (by
            :meth:`set_initial_model`)
        """
        if hasattr(self, "_initial_model") and self._initial_model is not None:
            return self._initial_model
        raise NotDefinedError(needs="initial_model")

    @property
    def model_shape(self) -> Union[Tuple, np.ndarray]:
        r"""the model shape

        Raises
        ------
        NotDefinedError
            when this property has not been defined (by either
            :meth:`set_model_shape` or :meth:`set_model_shape`)
        """
        if hasattr(self, "_model_shape") and self._model_shape is not None:
            return self._model_shape
        raise NotDefinedError(needs="model_shape")

    @property
    def blobs_dtype(self) -> list:
        r"""the name and type for the blobs that
        :meth:`log_posterior_with_blobs` will return

        Raises
        ------
        NotDefinedError
            when this property has not been defined (by either
            :meth:`set_blobs_dtype` or
            :meth:`set_log_posterior_with_blobs`)
        """
        if hasattr(self, "_blobs_dtype") and self._blobs_dtype is not None:
            return self._blobs_dtype
        raise NotDefinedError(needs="blobs name and type")

    @property
    def bounds(self):
        r"""TODO: document me!

        Raises
        ------
        NotDefinedError
            when this property has not been defined (by
            :meth:`set_bounds`)
        """
        if hasattr(self, "_bounds") and self._bounds is not None:
            return self._bounds
        raise NotDefinedError(needs="bounds")

    @property
    def constraints(self):
        r"""TODO: document me!

        Raises
        ------
        NotDefinedError
            when this property has not been defined (by
            :meth:`set_constraints`)
        """
        if hasattr(self, "_constraints") and self._constraints is not None:
            return self._constraints
        raise NotDefinedError(needs="constraints")

    @property
    def objective_defined(self) -> bool:
        r"""indicates whether :meth:`objective` has been defined"""
        return self._check_defined(self.objective)

    @property
    def log_posterior_defined(self) -> bool:
        r"""indicates whether :meth:`log_posterior` has been defined"""
        return self._check_defined(self.log_posterior)

    @property
    def log_posterior_with_blobs_defined(self) -> bool:
        r"""indicates whether :meth:`log_posterior_with_blobs` has been
        defined
        """
        return self._check_defined(self.log_posterior_with_blobs)

    @property
    def log_prior_defined(self) -> bool:
        r"""indicates whether :meth:`log_prior` has been defined"""
        return self._check_defined(self.log_prior)

    @property
    def log_likelihood_defined(self) -> bool:
        r"""indicates whether :meth:`log_likelihood` has been defined"""
        return self._check_defined(self.log_likelihood)

    @property
    def gradient_defined(self) -> bool:
        r"""indicates whether :meth:`gradient` has been defined"""
        return self._check_defined(self.gradient)

    @property
    def hessian_defined(self) -> bool:
        r"""indicates whether :meth:`hessian` has been defined"""
        return self._check_defined(self.hessian)

    @property
    def hessian_times_vector_defined(self) -> bool:
        r"""indicates whether :meth:`hessian_times_vector` has been defined"""
        return self._check_defined(self.hessian_times_vector, 2)

    @property
    def residual_defined(self) -> bool:
        r"""indicates whether :meth:`residual` has been defined"""
        return self._check_defined(self.residual)

    @property
    def jacobian_defined(self) -> bool:
        r"""indicates whether :meth:`jacobian` has been defined"""
        return self._check_defined(self.jacobian)

    @property
    def jacobian_times_vector_defined(self) -> bool:
        r"""indicates whether :meth:`jacobian_times_vector` has been defined"""
        return self._check_defined(self.jacobian_times_vector, 2)

    @property
    def data_misfit_defined(self) -> bool:
        r"""indicates whether :meth:`data_misfit` has been defined"""
        return self._check_defined(self.data_misfit)

    @property
    def regularization_defined(self) -> bool:
        r"""indicates whether :meth:`regularization` has been defined"""
        return self._check_defined(self.regularization)

    @property
    def regularization_matrix_defined(self) -> bool:
        r"""indicates whether :meth:`regularization_matrix` has been defined"""
        return self._check_defined(self.regularization_matrix)

    @property
    def forward_defined(self) -> bool:
        r"""indicates whether :meth:`forward` has been defined"""
        return self._check_defined(self.forward)

    @property
    def data_defined(self) -> bool:
        r"""indicates whether :meth:`data` has been defined"""
        return self._check_property_defined("data")

    @property
    def data_covariance_defined(self) -> bool:
        r"""indicates whether :meth:`data_covariance` has been defined"""
        return self._check_property_defined("data_covariance")

    @property
    def data_covariance_inv_defined(self) -> bool:
        r"""indicates whether :meth:`data_covariance_inv` has been defined"""
        return self._check_property_defined("data_covariance_inv")

    @property
    def initial_model_defined(self) -> bool:
        r"""indicates whether :meth:`initial_model` has been defined"""
        return self._check_property_defined("initial_model")

    @property
    def model_shape_defined(self) -> bool:
        r"""indicates whether :meth:`model_shape` has been defined"""
        return self._check_property_defined("model_shape")

    @property
    def blobs_dtype_defined(self) -> bool:
        r"""indicates whether :meth:`blobs_dtype` has been defined"""
        return self._check_property_defined("blobs_dtype")

    @property
    def bounds_defined(self) -> bool:
        r"""indicates whether :meth:`bounds` has been defined"""
        return self._check_property_defined("bounds")

    @property
    def constraints_defined(self) -> bool:
        r"""indicates whether :meth:`constraints` has been defined"""
        return self._check_property_defined("constraints")

    @staticmethod
    def _check_defined(func, args_num=1):
        if func is None:
            return False
        if isinstance(func, _FunctionWrapper):
            return True
        try:
            func(*[np.array([])] * args_num)
        except NotDefinedError:
            return False
        except Exception:  # ok if there're errors caused by dummy input
            return True
        else:
            return True  # ok if the function works without a wrapper

    def _check_property_defined(self, prop):
        try:
            getattr(self, prop)
        except NotDefinedError:
            return False
        else:
            return True

    # autogen_table: (tuple of defined things) ->
    #       (name of deduced item, reference to new function to generate)
    @property
    def autogen_table(self):
        return {
            ("data_misfit",): ("objective", _objective_from_dm),
            (
                "data_misfit",
                "regularization",
            ): ("objective", _objective_from_dm_reg),
            (
                "log_likelihood",
                "log_prior",
            ): ("log_posterior_with_blobs", _log_posterior_with_blobs_from_ll_lp),
            ("log_posterior_with_blobs",): (
                "log_posterior",
                _log_posterior_from_lp_with_blobs,
            ),
            ("hessian",): ("hessian_times_vector", _hessian_times_vector_from_hess),
            (
                "forward",
                "data",
            ): ("residual", _residual_from_fwd_dt),
            ("jacobian",): ("jacobian_times_vector", _jacobian_times_vector_from_jcb),
        }

    def _update_autogen(self, updated_item):
        update_dict = {k: v for k, v in self.autogen_table.items() if updated_item in k}
        for need_defined, (to_update, new_func) in update_dict.items():
            if getattr(self, f"{to_update}_defined"):
                to_update_existing = getattr(self, to_update)
                if (
                    isinstance(to_update_existing, _FunctionWrapper)
                    and not to_update_existing.autogen
                ):
                    continue  # already defined by user, don't overwrite
            if all(
                (getattr(self, f"{nm}_defined") for nm in need_defined)
            ):  # can update
                defined_items = list((getattr(self, nm) for nm in need_defined))
                new_func = _FunctionWrapper(
                    to_update, new_func, args=defined_items, autogen=True
                )
                setattr(self, to_update, new_func)
                if to_update == "log_posterior_with_blobs":
                    self.set_blobs_dtype(
                        [("log_likelihood", float), ("log_prior", float)]
                    )
                self._update_autogen(to_update)

    # ---------- Extra inforamtion inferred, not generally used by inference tools ----
    def model_covariance(self, model: np.ndarray):
        C_minv = self.model_covariance_inv(model)
        return np.linalg.inv(C_minv)

    def model_covariance_inv(self, model: np.ndarray):
        G = self.jacobian(model)
        C_dinv = self.data_covariance_inv
        return G.T @ C_dinv @ G

    # ---------- Display related ------------------------------------------------------
    @property
    def name(self) -> str:
        """Name of the current BaseProblem object, for display purposes, no actual
        meaning

        Returns
        -------
        str
            a name you've set
        """
        return self._name if hasattr(self, "_name") else self.__class__.__name__

    @name.setter
    def name(self, problem_name):
        self._name = problem_name

    def _data_misfit_squared_error(self, model: np.ndarray) -> Number:
        try:
            res = self.residual(model)
            if self.data_covariance_inv_defined:
                if _is_diag(self.data_covariance_inv):
                    weighted_res = np.diag(self.data_covariance_inv) * res
                    return res @ weighted_res
                else:
                    return res.T @ self.data_covariance_inv @ res
            elif self.data_covariance_defined and _is_diag(self.data_covariance):
                weighted_res = res / np.diag(self.data_covariance)
                return res @ weighted_res
            else:
                return np.sum(np.square(res))
        except Exception as exception:
            raise InvocationError(func_name="data misfit", autogen=True) from exception

    def summary(self):
        r"""Helper method that prints a summary of current ``BaseProblem`` object to
        console

        Examples
        --------

        .. admonition:: examples usage for BaseProblem.summary()
            :class: dropdown, attention

            .. code-block:: pycon
                :emphasize-lines: 6

                >>> from cofi import BaseProblem
                >>> import numpy as np
                >>> inv_problem = BaseProblem()
                >>> inv_problem.set_initial_model(np.array([1,2,3]))
                >>> inv_problem.set_data_misfit("L2")
                >>> inv_problem.summary()
                =====================================================================
                Summary for inversion problem: BaseProblem
                =====================================================================
                Model shape: (3,)
                ---------------------------------------------------------------------
                List of functions/properties set by you:
                ['initial_model', 'model_shape']
                ---------------------------------------------------------------------
                List of functions/properties created based on what you have provided:
                ['objective', 'data_misfit']
                ( Note that you did not set regularization )
                ---------------------------------------------------------------------
                List of functions/properties not set by you:
                (not all of these may be relevant to your inversion workflow)
                ['objective', 'gradient', 'hessian', 'hessian_times_vector', 'residual', 'jacobian', 'jacobian_times_vector', 'data_misfit', 'regularization', 'forward', 'data', 'bounds', 'constraints']

        """
        self._summary()

    def _summary(self, display_lines=True):
        # inspiration from keras: https://keras.io/examples/vision/mnist_convnet/
        title = f"Summary for inversion problem: {self.name}"
        sub_title1 = "List of functions/properties set by you:"
        sub_title2 = (
            "List of functions/properties created based on what you have provided:"
        )
        sub_title3 = (
            "List of functions/properties that can be further set for the problem:"
        )
        display_width = max(len(title), len(sub_title1), len(sub_title2))
        double_line = "=" * display_width
        single_line = "-" * display_width
        set_by_user, created_for_user = self._defined_components(False)
        not_set = [
            component
            for component in self.all_components
            if component not in set_by_user
        ]
        if display_lines:
            print(double_line)
        print(title)
        if display_lines:
            print(double_line)
        model_shape = self.model_shape if self.model_shape_defined else "Unknown"
        print(f"Model shape: {model_shape}")
        if display_lines:
            print(single_line)
        print(sub_title1)
        print(set_by_user if set_by_user else "-- none --")
        if display_lines:
            print(single_line)
        print(sub_title2)
        print(created_for_user if created_for_user else "-- none --")
        if (
            "objective" in created_for_user
            and self.data_misfit_defined
            and not self.regularization_defined
        ):
            print("( Note that you did not set regularization )")
        if display_lines:
            print(single_line)
        print(sub_title3)
        print("( not all of these may be relevant to your inversion workflow )")
        print(not_set if not_set else "-- none --")

    def __repr__(self) -> str:
        return f"{self.name}"


# ---------- End of BaseProblem class -------------------------------------------------


# ---------- Auto generated functions -------------------------------------------------
def _objective_from_dm_reg(model, data_misfit, regularization):
    try:
        return data_misfit(model) + regularization(model)
    except Exception as exception:
        raise InvocationError(
            func_name="objective function from data misfit and regularization",
            autogen=True,
        ) from exception


def _objective_from_dm(model, data_misfit):
    try:
        return data_misfit(model)
    except Exception as exception:
        raise InvocationError(
            func_name="objective function from data misfit",
            autogen=True,
        ) from exception


def _log_posterior_with_blobs_from_ll_lp(model, log_likelihood, log_prior):
    try:
        lp = log_prior(model)
        if lp == float("-inf"):
            return lp, None, lp
        ll = log_likelihood(model)
        return ll + lp, ll, lp
    except Exception as exception:
        raise InvocationError(
            func_name="log posterior function from log likelihood and log prior",
            autogen=True,
        ) from exception


def _log_posterior_from_lp_with_blobs(model, log_posterior_with_blobs):
    try:
        return log_posterior_with_blobs(model)[0]
    except Exception as exception:
        raise InvocationError(
            func_name="log posterior function from log likelihood and log prior",
            autogen=True,
        ) from exception


def _hessian_times_vector_from_hess(model, vector, hessian):
    try:
        return np.squeeze(np.asarray(hessian(model) @ vector))
    except Exception as exception:
        raise InvocationError(
            func_name="hessian_times_vector function from given hessian function",
            autogen=True,
        ) from exception


def _residual_from_fwd_dt(model, forward, data):
    try:
        return forward(model) - data
    except Exception as exception:
        raise InvocationError(
            func_name="residual function from forward and data provided",
            autogen=True,
        ) from exception


def _jacobian_times_vector_from_jcb(model, vector, jacobian):
    try:
        return np.squeeze(np.asarray(jacobian(model) @ vector))
    except Exception as exception:
        raise InvocationError(
            func_name="jacobian_times_vector from given jacobian function",
            autogen=True,
        ) from exception


def _matrix_to_func(_, matrix):
    return matrix


def _is_diag(matrix):
    diag_elem = np.diag(matrix).copy()
    np.fill_diagonal(matrix, 0)
    out = (matrix == 0).all()
    np.fill_diagonal(matrix, diag_elem)
    return out


# ---------- function wrapper to help make things pickleable --------------------------
class _FunctionWrapper:
    def __init__(
        self, name, func, args: list = None, kwargs: dict = None, autogen=False
    ):
        if not callable(func):
            raise InvalidOptionError(
                name=f"{name} function",
                invalid_option="not-callable input",
                valid_options="functions that are callable",
            )
        self.name = name
        self.func = func
        self.args = list() if args is None else args
        self.kwargs = dict() if kwargs is None else kwargs
        self.autogen = autogen
        try:  # not every function has __name__
            functools.update_wrapper(self, func)
            self.__name__ = func.__name__
            self.__doc__ = func.__doc__
        except:
            pass

    def __call__(self, model, *extra_args):
        try:
            return self.func(model, *extra_args, *self.args, **self.kwargs)
        except Exception as e:
            import traceback

            print(f"cofi: Exception while calling your {self.name} function:")
            print("  params:", model, *extra_args)
            print("  args:", self.args, len(self.args))
            print("  kwargs:", self.kwargs)
            print("  Look at details below:\n", "-" * 88)
            if self.autogen:
                raise
            else:
                raise InvocationError(func_name=self.name, autogen=self.autogen) from e
